import logging
import math
import re
from datetime import datetime

import pandas
from lxml import etree
from PySide2.QtCore import QObject, Signal, Slot
from PySide2.QtSql import QSqlTableModel
from PySide2.QtWidgets import QDialog, QFileDialog, QMessageBox
from jal.constants import Setup, TransactionType, PredefinedAsset, PredefinedCategory, CorporateAction
from jal.db.helpers import executeSQL, readSQL, get_country_by_code
from jal.ui_custom.helpers import g_tr
from jal.ui.ui_add_asset_dlg import Ui_AddAssetDialog
from jal.ui.ui_select_account_dlg import Ui_SelectAccountDlg


#-----------------------------------------------------------------------------------------------------------------------
class ReportType:
    IBKR = 'IBKR flex-query (*.xml)'
    Quik = 'Quik HTML-report (*.htm)'


class IBKRCashOp:
    Dividend = 0
    TaxWithhold = 1
    DepositWithdrawal = 2
    Fee = 3
    Interest = 4


#-----------------------------------------------------------------------------------------------------------------------
class IBKR:
    TaxNotePattern = "^(.*) - (..) TAX$"
    DummyExchange = "VALUE"
    SpinOffPattern = "^(.*)\(.* SPINOFF +(\d+) +FOR +(\d+) +\(.*$"
    SplitPattern = "^.* SPLIT +(\d+) +FOR +(\d+) +\(.*$"

    AssetType = {
        'CASH': PredefinedAsset.Money,
        'STK':  PredefinedAsset.Stock,
        # 'BOND': PredefinedAsset.Bond,
        'OPT':  PredefinedAsset.Derivative,
        'FUT':  PredefinedAsset.Derivative
    }

    CorpAction = {
        'TC': CorporateAction.Merger,
        'SO': CorporateAction.SpinOff,
        'IC': CorporateAction.SymbolChange,
        'HI': CorporateAction.StockDividend,
        'FS': CorporateAction.Split
    }

    @staticmethod
    def flString(data, name, default_value, _caller):
        if name not in data.attrib:
            return default_value
        return data.attrib[name]

    @staticmethod
    def flNumber(data, name, default_value, _caller):
        if name not in data.attrib:
            return default_value
        try:
            value = float(data.attrib[name])
        except ValueError:
            return None
        return value

    @staticmethod
    def flTimestamp(data, name, default_value, _caller):
        if name not in data.attrib:
            return default_value
        time_str = data.attrib[name]
        try:
            if len(time_str) == 15:  # YYYYMMDD;HHMMSS
                return int(datetime.strptime(time_str, "%Y%m%d;%H%M%S").timestamp())
            elif len(time_str) == 8: # YYYYMMDD
                return int(datetime.strptime(time_str, "%Y%m%d").timestamp())
            else:
                return default_value
        except ValueError:
            logging.error(g_tr('StatementLoader', "Unsupported date/time format: ") + f"{data.attrib[name]}")
            return None

    @staticmethod
    def flAssetType(data, name, default_value, _caller):
        if name not in data.attrib:
            return default_value
        try:
            return IBKR.AssetType[data.attrib[name]]
        except KeyError:
            if default_value is not None:
                return default_value
            else:
                logging.error(g_tr('StatementLoader', "Asset type isn't supported: ") + f"{data.attrib[name]}")
                return None

    @staticmethod
    def flCorpActionType(data, name, default_value, _caller):
        if name not in data.attrib:
            return default_value
        try:
            return IBKR.CorpAction[data.attrib[name]]
        except KeyError:
            if default_value is not None:
                return default_value
            else:
                logging.error(g_tr('StatementLoader', "Corporate action isn't supported: ") + f"{data.attrib[name]}")
                return None

    @staticmethod
    def flCashOpType(data, name, default_value, _caller):
        operations = {
            'Dividends':                    IBKRCashOp.Dividend,
            'Payment In Lieu Of Dividends': IBKRCashOp.Dividend,
            'Withholding Tax':              IBKRCashOp.TaxWithhold,
            'Deposits/Withdrawals':         IBKRCashOp.DepositWithdrawal,
            'Other Fees':                   IBKRCashOp.Fee,
            'Commission Adjustments':       IBKRCashOp.Fee,
            'Broker Interest Paid':         IBKRCashOp.Fee,
            'Broker Interest Received':     IBKRCashOp.Interest
        }

        if name not in data.attrib:
            return default_value
        try:
            return operations[data.attrib[name]]
        except KeyError:
            if default_value is not None:
                return default_value
            else:
                logging.error(g_tr('StatementLoader', "Cash transaction isn't supported: ") + f"{data.attrib[name]}")
                return None

    @staticmethod
    def flAccount(data, name, default_value, caller):
        if name not in data.attrib:
            return default_value
        if data.tag == 'Trade' and IBKR.flAssetType(data, 'assetCategory', None, None) == PredefinedAsset.Money:
            if 'symbol' not in data.attrib:
                logging.error(g_tr('StatementLoader', "Can't get currencies for accounts: ") + f"{data}")
                return None
            if 'ibCommissionCurrency' not in data.attrib:
                logging.error(g_tr('StatementLoader', "Can't get account currency for fee account: ") + f"{data}")
                return None
            currencies = data.attrib['symbol'].split('.')
            currencies.append(data.attrib['ibCommissionCurrency'])
            accountIds = []
            for currency in currencies:
                account = caller.findAccountID(data.attrib[name], currency)
                if account is None:
                    return None
                accountIds.append(account)
            return accountIds
        if 'currency' not in data.attrib:
            if default_value is None:
                logging.error(g_tr('StatementLoader', "Can't get account currency for account: ") + f"{data}")
            return default_value
        return caller.findAccountID(data.attrib[name], data.attrib['currency'])

    @staticmethod
    def flAsset(data, name, default_value, caller):
        if name not in data.attrib:
            return default_value
        if data.attrib[name] == '':
            return default_value
        if data.tag == 'Trade' and IBKR.flAssetType(data, 'assetCategory', None, None) == PredefinedAsset.Money:
            currency_asset = default_value
            for currency in data.attrib['symbol'].split('.'):
                currency_asset = caller.findAssetID(currency)
            return currency_asset
        if data.tag == 'CorporateAction' and IBKR.flString(data, 'listingExchange', None, None) == IBKR.DummyExchange:
            if data.attrib[name].endswith('.OLD'):
                return caller.findAssetID(data.attrib[name][:-len('.OLD')])
        return caller.findAssetID(data.attrib[name])

#-----------------------------------------------------------------------------------------------------------------------
class Quik:
    ClientPattern = "^Код клиента: (.*)$"
    DateTime = 'Дата и время заключения сделки'
    TradeNumber = 'Номер сделки'
    Symbol = 'Код инструмента'
    Name = 'Краткое наименование инструмента'
    Type = 'Направление'
    Qty = 'Кол-во'
    Price = 'Цена'
    Amount = 'Объём'
    Coupon = 'НКД'
    SettleDate = 'Дата расчётов'
    Buy = 'Купля'
    Sell = 'Продажа'
    Fee = 'Комиссия Брокера'
    FeeEx = 'Суммарная комиссия ТС'    # This line is used in KIT Broker reports
    FeeEx1 = 'Комиссия за ИТС'         # Below 3 lines are used in Uralsib Borker reports
    FeeEx2 = 'Комиссия за организацию торговли'
    FeeEx3 = 'Клиринговая комиссия'
    Total = 'ИТОГО'


#-----------------------------------------------------------------------------------------------------------------------
# Strip white spaces from numbers imported form Quik html-report
def convert_amount(val):
    val = val.replace(' ', '')
    try:
        res = float(val)
    except ValueError:
        res = 0
    return res


def addNewAsset(db, symbol, name, asset_type, isin, data_source=-1):
    _ = executeSQL(db, "INSERT INTO assets(name, type_id, full_name, isin, src_id) "
                       "VALUES(:symbol, :type, :full_name, :isin, :data_src)",
                   [(":symbol", symbol), (":type", asset_type), (":full_name", name),
                    (":isin", isin), (":data_src", data_source)])
    db.commit()
    asset_id = readSQL(db, "SELECT id FROM assets WHERE name=:symbol", [(":symbol", symbol)])
    if asset_id is None:
        logging.error(g_tr('', "Failed to add new asset: "), + f"{symbol}")
    return asset_id


#-----------------------------------------------------------------------------------------------------------------------
class AddAssetDialog(QDialog, Ui_AddAssetDialog):
    def __init__(self, parent, db, symbol):
        QDialog.__init__(self)
        self.setupUi(self)
        self.db = db
        self.asset_id = None

        self.SymbolEdit.setText(symbol)

        self.type_model = QSqlTableModel(db=db)
        self.type_model.setTable('asset_types')
        self.type_model.select()
        self.TypeCombo.setModel(self.type_model)
        self.TypeCombo.setModelColumn(1)

        self.data_src_model = QSqlTableModel(db=db)
        self.data_src_model.setTable('data_sources')
        self.data_src_model.select()
        self.DataSrcCombo.setModel(self.data_src_model)
        self.DataSrcCombo.setModelColumn(1)

        # center dialog with respect to parent window
        x = parent.x() + parent.width()/2 - self.width()/2
        y = parent.y() + parent.height()/2 - self.height()/2
        self.setGeometry(x, y, self.width(), self.height())

    def accept(self):
        self.asset_id = addNewAsset(self.db, self.SymbolEdit.text(), self.NameEdit.text(),
                                    self.type_model.record(self.TypeCombo.currentIndex()).value("id"),
                                    self.isinEdit.text(),
                                    self.data_src_model.record(self.DataSrcCombo.currentIndex()).value("id"))
        super().accept()


#-----------------------------------------------------------------------------------------------------------------------
class SelectAccountDialog(QDialog, Ui_SelectAccountDlg):
    def __init__(self, parent, db, description, current_account):
        QDialog.__init__(self)
        self.setupUi(self)
        self.db = db
        self.account_id = None
        self.current_account = current_account

        self.DescriptionLbl.setText(description)
        self.AccountWidget.init_db(db)

        # center dialog with respect to parent window
        x = parent.x() + parent.width() / 2 - self.width() / 2
        y = parent.y() + parent.height() / 2 - self.height() / 2
        self.setGeometry(x, y, self.width(), self.height())

    @Slot()
    def closeEvent(self, event):
        self.account_id = self.AccountWidget.selected_id
        if self.AccountWidget.selected_id == 0:
            QMessageBox().warning(None, g_tr('ReferenceDataDialog', "No selection"),
                                  g_tr('ReferenceDataDialog', "Invalid account selected"),
                                  QMessageBox.Ok)
            event.ignore()
            return

        if self.AccountWidget.selected_id == self.current_account:
            QMessageBox().warning(None, g_tr('ReferenceDataDialog', "No selection"),
                                  g_tr('ReferenceDataDialog', "Please select different account"),
                                  QMessageBox.Ok)
            event.ignore()
            return

        event.accept()

#-----------------------------------------------------------------------------------------------------------------------
class StatementLoader(QObject):
    load_completed = Signal()
    load_failed = Signal()

    def __init__(self, parent, db):
        super().__init__()
        self.parent = parent
        self.db = db
        self.loaders = {
            ReportType.IBKR: self.loadIBFlex,
            ReportType.Quik: self.loadQuikHtml
        }
        self.currentIBstatement = None

    # Displays file choose dialog and loads corresponding report if user have chosen a file
    def loadReport(self):
        report_file, active_filter = \
            QFileDialog.getOpenFileName(None, g_tr('StatementLoader', "Select statement file to import"),
                                        ".", f"{ReportType.IBKR};;{ReportType.Quik}")
        if report_file:
            result = self.loaders[active_filter](report_file)
            if result:
                self.load_completed.emit()
            else:
                self.load_failed.emit()

    # Searches for account_id by account number and optional currency
    # Returns: account_id or None if no account was found
    def findAccountID(self, accountNumber, accountCurrency=''):
        if accountCurrency:
            account_id = readSQL(self.db, "SELECT a.id FROM accounts AS a "
                                          "LEFT JOIN assets AS c ON c.id=a.currency_id "
                                          "WHERE a.number=:account_number AND c.name=:currency_name",
                                 [(":account_number", accountNumber), (":currency_name", accountCurrency)])
        else:
            account_id = readSQL(self.db, "SELECT a.id FROM accounts AS a "
                                          "LEFT JOIN assets AS c ON c.id=a.currency_id "
                                          "WHERE a.number=:account_number", [(":account_number", accountNumber)])
        if account_id is None:
            logging.error(g_tr('StatementLoader', "Account not found: ") + f"{accountNumber} ({accountCurrency})")
        return account_id

    # Searches for asset_id in database and returns it.
    # If asset is not found - shows dialog for new asset creation.
    # Returns: asset_id or None if new asset creation failed
    def findAssetID(self, symbol):
        asset_id = readSQL(self.db, "SELECT id FROM assets WHERE name=:symbol", [(":symbol", symbol)])
        if asset_id is None:
            dialog = AddAssetDialog(self.parent, self.db, symbol)
            dialog.exec_()
            asset_id = dialog.asset_id
        return asset_id

    # returns bank id assigned for the account or asks for assignment if field is empty
    def getAccountBank(self, account_id):
        bank_id = readSQL(self.db, "SELECT organization_id FROM accounts WHERE id=:account_id",
                          [(":account_id", account_id)])
        if bank_id != '':
            return bank_id
        bank_id = readSQL(self.db, "SELECT id FROM agents WHERE name='Interactive Brokers'")
        if bank_id is not None:
            return bank_id
        query = executeSQL(self.db, "INSERT INTO agents (pid, name) VALUES (0, 'Interactive Brokers')")
        bank_id =query.lastInsertId()
        _ = executeSQL(self.db, "UPDATE accounts SET organization_id=:bank_id WHERE id=:account_id",
                       [(":bank_id", bank_id), (":account_id", account_id)])
        return bank_id

    def loadIBFlex(self, filename):
        section_loaders = {
            'SecuritiesInfo':   self.loadIBSecurities,    # Order of load is important - SecuritiesInfo is first
            'Trades':           self.loadIBTrades,
            'OptionEAE':        self.loadIBOptions,
            'CorporateActions': self.loadIBCorporateActions,
            'CashTransactions': self.loadIBCashTransactions,
            'TransactionTaxes': self.loadIBTaxes
        }
        try:
            xml_root = etree.parse(filename)
            for FlexStatements in xml_root.getroot():
                for statement in FlexStatements:
                    attr = statement.attrib
                    logging.info(g_tr('StatementLoader', "Load IB Flex-statement for account ") +
                                 f"{attr['accountId']}: {attr['fromDate']} - {attr['toDate']}")
                    for section in section_loaders:
                        section_elements = statement.xpath(section)  # Actually should be list of 0 or 1 element
                        if section_elements:
                            section_data = self.getIBdata(section_elements[0])
                            if section_data is None:
                                return False
                            section_loaders[section](section_data)
        except Exception as e:
            logging.error(g_tr('StatementLoader', "Failed to parse Interactive Brokers flex-report") + f": {e}")
            return False
        logging.info(g_tr('StatementLoader', "IB Flex-statement loaded successfully"))
        return True

    def getIBdata(self, section):
        section_descriptions = {
            'CashTransactions': {'tag': 'CashTransaction', 'values': [('type', IBKR.flCashOpType, None),
                                                                      ('accountId', IBKR.flAccount, None),
                                                                      ('currency', IBKR.flString, ''),
                                                                      ('symbol', IBKR.flAsset, 0),
                                                                      ('dateTime', IBKR.flTimestamp, None),
                                                                      ('amount', IBKR.flNumber, None),
                                                                      ('description', IBKR.flString, None)]},
            'CorporateActions': {'tag': 'CorporateAction', 'values': [('type', IBKR.flCorpActionType, None),
                                                                      ('accountId', IBKR.flAccount, None),
                                                                      ('symbol', IBKR.flAsset, None),
                                                                      ('listingExchange', IBKR.flString, ''),
                                                                      ('assetCategory', IBKR.flAssetType, None),
                                                                      ('dateTime', IBKR.flTimestamp, None),
                                                                      ('transactionID', IBKR.flString, ''),
                                                                      ('description', IBKR.flString, None),
                                                                      ('quantity', IBKR.flNumber, None),
                                                                      ('code', IBKR.flString, '')]},
            'OptionEAE': {'tag': 'OptionEAE', 'values': [('transactionType', IBKR.flString, None),
                                                         ('symbol', IBKR.flAsset, None),
                                                         ('accountId', IBKR.flAccount, None),
                                                         ('date', IBKR.flTimestamp, None),
                                                         ('tradePrice', IBKR.flNumber, None),
                                                         ('quantity', IBKR.flNumber, None),
                                                         ('multiplier', IBKR.flNumber, None),
                                                         ('commisionsAndTax', IBKR.flNumber, None),
                                                         ('tradeID', IBKR.flString, ''),
                                                         ('notes', IBKR.flString, '')]},
            'SecuritiesInfo': {'tag': 'SecurityInfo',
                               'values': [('symbol', IBKR.flString, None),
                                          ('assetCategory', IBKR.flAssetType, None),
                                          ('subCategory', IBKR.flString, ''),
                                          ('description', IBKR.flString, None),
                                          ('isin', IBKR.flString, '')]},
            'Trades': {'tag': 'Trade', 'values': [('assetCategory', IBKR.flAssetType, None),
                                                  ('symbol', IBKR.flAsset, None),
                                                  ('accountId', IBKR.flAccount, None),
                                                  ('dateTime', IBKR.flTimestamp, None),
                                                  ('settleDateTarget', IBKR.flTimestamp, 0),
                                                  ('tradePrice', IBKR.flNumber, None),
                                                  ('quantity', IBKR.flNumber, None),
                                                  ('proceeds', IBKR.flNumber, None),
                                                  ('multiplier', IBKR.flNumber, None),
                                                  ('ibCommission', IBKR.flNumber, None),
                                                  ('tradeID', IBKR.flString, ''),
                                                  ('exchange', IBKR.flString, ''),
                                                  ('notes', IBKR.flString, '')]},
            'TransactionTaxes': {'tag': 'TransactionTax', 'values': [('accountId', IBKR.flAccount, None),
                                                                     ('symbol', IBKR.flString, ''),
                                                                     ('date', IBKR.flTimestamp, None),
                                                                     ('taxAmount', IBKR.flNumber, None),
                                                                     ('description', IBKR.flString, None),
                                                                     ('taxDescription', IBKR.flString, None)]}
        }

        try:
            tag = section_descriptions[section.tag]['tag']
        except KeyError:
            return []           # This section isn't used for import

        data = []
        for sample in section.xpath(tag):
            tag_dictionary = {}
            for attr_name, attr_loader, attr_default in section_descriptions[section.tag]['values']:
                attr_value = attr_loader(sample, attr_name, attr_default, self)
                if attr_value is None:
                    logging.error(g_tr('StatementLoader', "Failed to load attribute: ") + f"{attr_name} / {sample.attrib}")
                    return None
                tag_dictionary[attr_name] = attr_value
            data.append(tag_dictionary)
        return data

    def loadIBSecurities(self, assets):
        cnt = 0
        for asset in assets:
            if readSQL(self.db, "SELECT id FROM assets WHERE name=:symbol", [(":symbol", asset['symbol'])]):
                continue
            asset_type = PredefinedAsset.ETF if asset['subCategory'] == "ETF" else asset['assetCategory']
            addNewAsset(self.db, asset['symbol'], asset['description'], asset_type, asset['isin'])
            cnt += 1
        logging.info(g_tr('StatementLoader', "Securities loaded: ") + f"{cnt} ({len(assets)})")

    def loadIBTrades(self, trades):
        ib_trade_loaders = {
            PredefinedAsset.Stock:      self.loadIBStockTrade,
            PredefinedAsset.Derivative: self.loadIBStockTrade,
            PredefinedAsset.Money:      self.loadIBCurrencyTrade
        }

        cnt = 0
        for trade in trades:
            cnt += ib_trade_loaders[trade['assetCategory']](trade)
        logging.info(g_tr('StatementLoader', "Trades loaded: ") + f"{cnt} ({len(trades)})")

    def loadIBOptions(self, options):
        ib_option_loaders = {
            "Assignment": self.loadIBOptionEAE,
            "Exercise":   self.loadIBOptionEAE,
            "Expiration": self.loadIBOptionEAE,
            "Buy":        self.loadIBStockTrade,
            "Sell":       self.loadIBStockTrade,
        }
        cnt = 0
        for option in options:
            try:
                option['dateTime'] = option['date']
                option['settleDateTarget'] = option['date']
                option['ibCommission'] = option['commisionsAndTax']
                cnt += ib_option_loaders[option['transactionType']](option)
            except KeyError:
                logging.error(
                    g_tr('StatementLoader', "Option E&A&E action isn't implemented: ") + f"{option['transactionType']}")
        logging.info(g_tr('StatementLoader', "Options E&A&E loaded: ") + f"{cnt} ({len(options)})")

    def loadIBCorporateActions(self, actions):
        cnt = 0
        for action in actions:
            if action['listingExchange'] == IBKR.DummyExchange:  # Skip actions that we loaded as part of main action
                continue
            if action['code'] == 'Ca':
                logging.warning(g_tr('StatementLoader', "*** MANUAL ACTION REQUIRED ***"))
                logging.warning(g_tr('StatementLoader', "Corporate action cancelled: ") + f"{action}")
                continue
            if action['assetCategory'] != PredefinedAsset.Stock:
                logging.warning(g_tr('StatementLoader', "Corporate actions are supported for stocks only"))
                continue

            if action['type'] == CorporateAction.Merger:
                # additional info is in previous dummy record where original symbol and quantity are present
                pair_idu = str(int(action['transactionID']) + 1)
                pair_idl = str(int(action['transactionID']) - 1)
                paired_records = list(filter(
                    lambda pair: (pair['transactionID'] == pair_idl or pair['transactionID'] == pair_idu)
                                 and pair['listingExchange'] == IBKR.DummyExchange
                                 and pair['type'] == action['type']
                                 and pair['description'][:15] == action['description'][:15], actions))
                if len(paired_records) != 1:
                    logging.error(g_tr('StatementLoader', "Can't find paired record for ") + f"{action}")
                    continue
                self.createCorpAction(action['accountId'], CorporateAction.Merger, action['dateTime'], action['transactionID'], paired_records[0]['symbol'],
                                      -paired_records[0]['quantity'], action['symbol'], action['quantity'], action['description'])
                cnt += 2
            elif action['type'] == CorporateAction.SpinOff:
                parts = re.match(IBKR.SpinOffPattern, action['description'], re.IGNORECASE)
                if not parts:
                    logging.error(g_tr('StatementLoader', "Failed to parse Spin-off data for ") + f"{action}")
                    continue
                asset_id_old = self.findAssetID(parts.group(1))
                mult_a = int(parts.group(2))
                mult_b = int(parts.group(3))
                qty_old = mult_b * action['quantity'] / mult_a
                self.createCorpAction(action['accountId'], CorporateAction.SpinOff, action['dateTime'], action['transactionID'], asset_id_old,
                                      qty_old, action['symbol'], action['quantity'], action['description'])
                cnt += 1
            elif action['type'] == CorporateAction.SymbolChange:
                # additional info is in next dummy record where old symbol is changed to *.OLD
                pair_idu = str(int(action['transactionID']) + 1)
                pair_idl = str(int(action['transactionID']) - 1)
                paired_records = list(filter(
                    lambda pair: (pair['transactionID'] == pair_idl or pair['transactionID'] == pair_idu)
                                 and pair['listingExchange'] == IBKR.DummyExchange
                                 and pair['type'] == action['type']
                                 and pair['description'][:15] == action['description'][:15], actions))
                if len(paired_records) != 1:
                    logging.error(g_tr('StatementLoader', "Can't find paired record for: ") + f"{action}")
                    continue
                self.createCorpAction(action['accountId'], CorporateAction.SymbolChange, action['dateTime'], action['transactionID'], paired_records[0]['symbol'],
                                      action['quantity'], action['symbol'], action['quantity'], action['description'])
                cnt += 2
            elif action['type'] == CorporateAction.StockDividend:
                self.createCorpAction(action['accountId'], CorporateAction.StockDividend, action['dateTime'], action['transactionID'], action['symbol'], 0,
                                      action['symbol'], action['quantity'], action['description'])
                cnt += 1
            elif action['type'] == CorporateAction.Split:
                parts = re.match(IBKR.SplitPattern, action['description'], re.IGNORECASE)
                if not parts:
                    logging.error(g_tr('StatementLoader', "Failed to parse corp.action Split data"))
                    return
                mult_a = int(parts.group(1))
                mult_b = int(parts.group(2))
                qty_new = mult_a * action['quantity'] / mult_b
                self.createCorpAction(action['accountId'], CorporateAction.Split, action['dateTime'], action['transactionID'], action['symbol'],
                                      action['quantity'], action['symbol'], qty_new, action['description'])
                cnt += 1
            else:
                logging.error(g_tr('StatementLoader', "Corporate action type is not supported: ")
                              + f"{action['type']}")
                continue
        logging.info(g_tr('StatementLoader', "Corporate actions loaded: ") + f"{cnt} ({len(actions)})")

    def loadIBTaxes(self, taxes):
        cnt = 0
        for tax in taxes:
            bank_id = self.getAccountBank(tax['accountId'])
            note = f"{tax['symbol']} ({tax['description']}) - {tax['taxDescription']}"
            id = readSQL(self.db, "SELECT id FROM all_operations WHERE type = :type "
                                  "AND timestamp=:timestamp AND account_id=:account_id AND amount=:amount",
                         [(":timestamp", tax['date']), (":type", TransactionType.Action),
                          (":account_id", tax['accountId']), (":amount", tax['taxAmount'])])
            if id:
                logging.warning(g_tr('StatementLoader', "Tax transaction already exists ") + f"{tax}")
                continue
            query = executeSQL(self.db, "INSERT INTO actions (timestamp, account_id, peer_id) "
                                        "VALUES (:timestamp, :account_id, :bank_id)",
                               [(":timestamp", tax['date']), (":account_id", tax['accountId']), (":bank_id", bank_id)])
            pid = query.lastInsertId()
            _ = executeSQL(self.db, "INSERT INTO action_details (pid, category_id, sum, note) "
                                    "VALUES (:pid, :category_id, :sum, :note)",
                           [(":pid", pid), (":category_id", PredefinedCategory.Taxes),
                            (":sum", tax['taxAmount']), (":note", note)])
            self.db.commit()
            cnt += 1
        logging.info(g_tr('StatementLoader', "Taxes loaded: ") + f"{cnt} ({len(taxes)})")

    def loadIBCashTransactions(self, cash):
        cnt = 0

        dividends = list(filter(lambda tr: tr['type'] == IBKRCashOp.Dividend, cash))
        for dividend in dividends:
            cnt += self.loadIBDividend(dividend)

        taxes = list(filter(lambda tr: tr['type'] == IBKRCashOp.TaxWithhold, cash))
        for tax in taxes:
            cnt += self.loadIBWithholdingTax(tax)

        transfers = list(filter(lambda tr: tr['type'] == IBKRCashOp.DepositWithdrawal, cash))
        for transfer in transfers:
            cnt += self.loadIBDepositWithdraw(transfer)

        fees = list(filter(lambda tr: tr['type'] == IBKRCashOp.Fee, cash))
        for fee in fees:
            cnt += self.loadIBFee(fee)

        interests = list(filter(lambda tr: tr['type'] == IBKRCashOp.Interest, cash))
        for interest in interests:
            cnt += self.loadIBInterest(interest)

        logging.info(g_tr('StatementLoader', "Cash transactions loaded: ") + f"{cnt} ({len(cash)})")

    def loadIBStockTrade(self, trade):
        qty = trade['quantity'] * trade['multiplier']
        if trade['settleDateTarget'] == 0:
            trade['settleDateTarget'] = trade['dateTime']
        if trade['notes'] == 'Ca':
            self.deleteTrade(trade['accountId'], trade['symbol'], trade['dateTime'], trade['settleDateTarget'],
                             trade['tradeID'], qty, trade['tradePrice'], trade['ibCommission'])
        else:
            self.createTrade(trade['accountId'], trade['symbol'], trade['dateTime'], trade['settleDateTarget'],
                             trade['tradeID'], qty, trade['tradePrice'], trade['ibCommission'])
        return 1

    def loadIBOptionEAE(self, option):
        qty = option['quantity'] * option['multiplier']
        self.createTrade(option['accountId'], option['symbol'], option['date'], option['date'], option['tradeID'], qty,
                         option['tradePrice'], option['commisionsAndTax'])
        return 1

    def createTrade(self, account_id, asset_id, timestamp, settlement, number, qty, price, fee, coupon=0.0):
        trade_id = readSQL(self.db,
                           "SELECT id FROM trades "
                           "WHERE timestamp=:timestamp AND asset_id = :asset "
                           "AND account_id = :account AND number = :number AND qty = :qty AND price = :price",
                           [(":timestamp", timestamp), (":asset", asset_id), (":account", account_id),
                            (":number", number), (":qty", qty), (":price", price)])
        if trade_id:
            logging.info(g_tr('StatementLoader', "Trade already exists: #") + f"{number}")
            return

        _ = executeSQL(self.db,
                       "INSERT INTO trades (timestamp, settlement, number, account_id, "
                       "asset_id, qty, price, fee, coupon) "
                       "VALUES (:timestamp, :settlement, :number, :account, :asset, :qty, :price, :fee, :coupon)",
                       [(":timestamp", timestamp), (":settlement", settlement), (":number", number),
                        (":account", account_id), (":asset", asset_id), (":qty", float(qty)),
                        (":price", float(price)), (":fee", -float(fee)), (":coupon", float(coupon))])
        self.db.commit()

    def deleteTrade(self, account_id, asset_id, timestamp, _settlement, number, qty, price, _fee):
        _ = executeSQL(self.db, "DELETE FROM trades "
                                "WHERE timestamp=:timestamp AND asset_id=:asset "
                                "AND account_id=:account AND number=:number AND qty=:qty AND price=:price",
                       [(":timestamp", timestamp), (":asset", asset_id), (":account", account_id),
                        (":number", number), (":qty", -qty), (":price", price)])
        self.db.commit()

    def loadIBCurrencyTrade(self, trade):
        if trade['quantity'] > 0:
            from_idx = 1
            to_idx = 0
            to_amount = trade['quantity']  # positive value
            from_amount = trade['proceeds']  # already negative value
        elif trade['quantity'] < 0:
            from_idx = 0
            to_idx = 1
            from_amount = trade['quantity']  # already negative value
            to_amount = trade['proceeds']  # positive value
        else:
            logging.error(g_tr('StatementLoader', "Zero quantity in cash trade: ") + f"{trade}")
            return 0
        self.createTransfer(trade['dateTime'], trade['accountId'][from_idx], from_amount,
                            trade['accountId'][to_idx], to_amount, trade['accountId'][2], trade['ibCommission'], trade['exchange'])
        return 1

    def createTransfer(self, timestamp, f_acc_id, f_amount, t_acc_id, t_amount, fee_acc_id, fee, note):
        transfer_id = readSQL(self.db,
                              "SELECT id FROM transfers_combined "
                              "WHERE from_timestamp=:timestamp AND from_acc_id=:from_acc_id AND to_acc_id=:to_acc_id",
                              [(":timestamp", timestamp), (":from_acc_id", f_acc_id), (":to_acc_id", t_acc_id)])
        if transfer_id:
            logging.info(g_tr('StatementLoader', "Transfer/Exchange already exists: ") + f"{f_amount}->{t_amount}")
            return
        if abs(fee) > Setup.CALC_TOLERANCE:
            _ = executeSQL(self.db,
                           "INSERT INTO transfers_combined (from_timestamp, from_acc_id, from_amount, "
                           "to_timestamp, to_acc_id, to_amount, fee_timestamp, fee_acc_id, fee_amount, note) "
                           "VALUES (:timestamp, :f_acc_id, :f_amount, :timestamp, :t_acc_id, :t_amount, "
                           ":timestamp, :fee_acc_id, :fee_amount, :note)",
                           [(":timestamp", timestamp), (":f_acc_id", f_acc_id), (":t_acc_id", t_acc_id),
                            (":f_amount", f_amount), (":t_amount", t_amount), (":fee_acc_id", fee_acc_id),
                            (":fee_amount", fee), (":note", note)])
        else:
            _ = executeSQL(self.db,
                           "INSERT INTO transfers_combined (from_timestamp, from_acc_id, from_amount, "
                           "to_timestamp, to_acc_id, to_amount, note) "
                           "VALUES (:timestamp, :f_acc_id, :f_amount, :timestamp, :t_acc_id, :t_amount, :note)",
                           [(":timestamp", timestamp), (":f_acc_id", f_acc_id), (":t_acc_id", t_acc_id),
                            (":f_amount", f_amount), (":t_amount", t_amount), (":note", note)])
        self.db.commit()

    def createCorpAction(self, account_id, type, timestamp, number, asset_id_old, qty_old, asset_id_new, qty_new, note):
        action_id = readSQL(self.db,
                           "SELECT id FROM corp_actions "
                           "WHERE timestamp=:timestamp AND type = :type AND account_id = :account AND number = :number "
                           "AND asset_id = :asset AND asset_id_new = :asset_new",
                           [(":timestamp", timestamp), (":type", type), (":account", account_id), (":number", number),
                            (":asset", asset_id_old), (":asset_new", asset_id_new)])
        if action_id:
            logging.info(g_tr('StatementLoader', "Corporate action already exists: #") + f"{number}")
            return

        _ = executeSQL(self.db,
                       "INSERT INTO corp_actions (timestamp, number, account_id, type, "
                       "asset_id, qty, asset_id_new, qty_new, note) "
                       "VALUES (:timestamp, :number, :account, :type, :asset, :qty, :asset_new, :qty_new, :note)",
                       [(":timestamp", timestamp), (":number", number), (":account", account_id), (":type", type),
                        (":asset", asset_id_old), (":qty", float(qty_old)),
                        (":asset_new", asset_id_new), (":qty_new", float(qty_new)), (":note", note)])
        self.db.commit()

    def loadIBDividend(self, dividend):
        self.createDividend(dividend['dateTime'], dividend['accountId'], dividend['symbol'], dividend['amount'], dividend['description'])
        return 1

    def loadIBWithholdingTax(self, tax):
        self.addWithholdingTax(tax['dateTime'], tax['accountId'], tax['symbol'], -tax['amount'], tax['description'])
        return 1

    def loadIBFee(self, fee):
        bank_id = self.getAccountBank(fee['accountId'])
        query = executeSQL(self.db, "INSERT INTO actions (timestamp, account_id, peer_id) "
                                    "VALUES (:timestamp, :account_id, :bank_id)",
                           [(":timestamp", fee['dateTime']), (":account_id", fee['accountId']), (":bank_id", bank_id)])
        pid = query.lastInsertId()
        _ = executeSQL(self.db, "INSERT INTO action_details (pid, category_id, sum, note) "
                                "VALUES (:pid, :category_id, :sum, :note)",
                       [(":pid", pid), (":category_id", PredefinedCategory.Fees), (":sum", fee['amount']),
                        (":note", fee['description'])])
        self.db.commit()
        return 1

    def loadIBInterest(self, interest):
        bank_id = self.getAccountBank(interest['accountId'])
        query = executeSQL(self.db, "INSERT INTO actions (timestamp, account_id, peer_id) "
                                    "VALUES (:timestamp, :account_id, :bank_id)",
                           [(":timestamp", interest['dateTime']), (":account_id", interest['accountId']),
                            (":bank_id", bank_id)])
        pid = query.lastInsertId()
        _ = executeSQL(self.db, "INSERT INTO action_details (pid, category_id, sum, note) "
                                "VALUES (:pid, :category_id, :sum, :note)",
                       [(":pid", pid), (":category_id", PredefinedCategory.Interest), (":sum", interest['amount']),
                        (":note", interest['description'])])
        self.db.commit()
        return 1

    # noinspection PyMethodMayBeStatic
    def loadIBDepositWithdraw(self, cash):
        if cash['amount'] >= 0:     # Deposit
            text = g_tr('StatementLoader', "Deposit of ") + f"{cash['amount']:.2f} {cash['currency']} " + \
                   f"@{datetime.fromtimestamp(cash['dateTime']).strftime('%d.%m.%Y')}\n" + \
                   g_tr('StatementLoader', "Select account to withdraw from:")
        else:                       # Withdrawal
            text = g_tr('StatementLoader', "Withdrawal of ") + f"{-cash['amount']:.2f} {cash['currency']} " + \
                   f"@{datetime.fromtimestamp(cash['dateTime']).strftime('%d.%m.%Y')}\n" + \
                   g_tr('StatementLoader', "Select account to deposit to:")

        dialog = SelectAccountDialog(self.parent, self.db, text, cash['accountId'])
        if dialog.exec_() != QDialog.Accepted:
            return 0

        if cash['amount'] >= 0:
            self.createTransfer(cash['dateTime'], dialog.account_id, -cash['amount'],
                                cash['accountId'], cash['amount'], 0, 0, cash['description'])
        else:
            self.createTransfer(cash['dateTime'], cash['accountId'], cash['amount'],
                                dialog.account_id, -cash['amount'], 0, 0, cash['description'])
        return 1

    def createDividend(self, timestamp, account_id, asset_id, amount, note):
        id = readSQL(self.db, "SELECT id FROM dividends WHERE timestamp=:timestamp "
                              "AND account_id=:account_id AND asset_id=:asset_id AND note=:note",
                     [(":timestamp", timestamp), (":account_id", account_id), (":asset_id", asset_id), (":note", note)])
        if id:
            logging.info(g_tr('StatementLoader', "Dividend already exists: ") + f"{note}")
            return
        _ = executeSQL(self.db, "INSERT INTO dividends (timestamp, account_id, asset_id, sum, note) "
                                "VALUES (:timestamp, :account_id, :asset_id, :sum, :note)",
                       [(":timestamp", timestamp), (":account_id", account_id), (":asset_id", asset_id),
                        (":sum", amount), (":note", note)])
        self.db.commit()

    def addWithholdingTax(self, timestamp, account_id, asset_id, amount, note):
        parts = re.match(IBKR.TaxNotePattern, note)
        if not parts:
            logging.warning(g_tr('StatementLoader', "*** MANUAL ENTRY REQUIRED ***"))
            logging.warning(g_tr('StatementLoader', "Unhandled tax pattern found: ") + f"{note}")
            return
        dividend_note = parts.group(1) + '%'
        country_code = parts.group(2).lower()
        country_id = get_country_by_code(self.db, country_code)
        if country_id == 0:
            query = executeSQL(self.db, "INSERT INTO countries(name, code, tax_treaty) VALUES (:name, :code, 0)",
                               [(":name", "Country_" + country_code), (":code", country_code)])
            country_id = query.lastInsertId()
            logging.warning(g_tr('StatementLoader', "New country added (set Tax Treaty in Data->Countries menu): ")
                            + f"'{country_code}'")
        try:
            dividend_id, old_tax = readSQL(self.db,
                                           "SELECT id, sum_tax FROM dividends "
                                           "WHERE timestamp=:timestamp AND account_id=:account_id "
                                           "AND asset_id=:asset_id AND note LIKE :dividend_description",
                                           [(":timestamp", timestamp), (":account_id", account_id),
                                            (":asset_id", asset_id), (":dividend_description", dividend_note)])
        except:
            logging.warning(g_tr('StatementLoader', "Dividend not found for withholding tax: ") + f"{note}")
            return
        _ = executeSQL(self.db, "UPDATE dividends SET sum_tax=:tax, tax_country_id=:country_id WHERE id=:dividend_id",
                       [(":dividend_id", dividend_id), (":tax", old_tax + amount), (":country_id", country_id)])
        self.db.commit()

    def loadQuikHtml(self, filename):
        try:
            data = pandas.read_html(filename, encoding='cp1251',
                                    converters={Quik.Qty: convert_amount, Quik.Amount: convert_amount,
                                                Quik.Price: convert_amount, Quik.Coupon: convert_amount})
        except:
            logging.error(g_tr('StatementLoader', "Can't read statement file"))
            return False

        report_info = data[0]
        deals_info = data[1]
        parts = re.match(Quik.ClientPattern, report_info[0][2])
        if parts:
            account_id = self.findAccountID(parts.group(1))
        else:
            logging.error(g_tr('StatementLoader', "Can't get account number from the statement."))
            return False
        if account_id is None:
            logging.error(g_tr('StatementLoader', "Account with number ") + f"{parts.group(1)}" +
                          g_tr('StatementLoader', " not found. Import cancelled."))
            return False

        for index, row in deals_info.iterrows():
            if row[Quik.Type] == Quik.Buy:
                qty = int(row[Quik.Qty])
            elif row[Quik.Type] == Quik.Sell:
                qty = -int(row[Quik.Qty])
            elif row[Quik.Type][:len(Quik.Total)] == Quik.Total:
                break   # End of statement reached
            else:
                logging.warning(g_tr('StatementLoader', "Unknown operation type ") + f"'{row[Quik.Type]}'")
                continue
            asset_id = self.findAssetID(row[Quik.Symbol])
            if asset_id is None:
                logging.warning(g_tr('StatementLoader', "Unknown asset ") + f"'{row[Quik.Symbol]}'")
                continue
            timestamp = int(datetime.strptime(row[Quik.DateTime], "%d.%m.%Y %H:%M:%S").timestamp())
            settlement = int(datetime.strptime(row[Quik.SettleDate], "%d.%m.%Y").timestamp())
            number = row[Quik.TradeNumber]
            price = row[Quik.Price]
            amount = row[Quik.Amount]
            lot_size = math.pow(10, round(math.log10(amount / (price * abs(qty)))))
            qty = qty * lot_size
            fee = float(row[Quik.Fee])
            if Quik.FeeEx in row:  # Broker dependent fee import
                fee = fee + float(row[Quik.FeeEx])
            else:
                fee = fee + float(row[Quik.FeeEx1]) + float(row[Quik.FeeEx2]) + float(row[Quik.FeeEx3])
            coupon = float(row[Quik.Coupon])
            self.createTrade(account_id, asset_id, timestamp, settlement, number, qty, price, -fee, coupon)
        return True


import sys
import os
import logging
import traceback
from PySide2.QtCore import QTranslator
from PySide2.QtWidgets import QApplication
from jal.widgets.main_window import MainWindow, AbortWindow
from jal.db.helpers import init_and_check_db, LedgerInitError, get_language


#-----------------------------------------------------------------------------------------------------------------------
def exception_logger(exctype, value, tb):
    info = traceback.format_exception(exctype, value, tb)
    logging.fatal(f"EXCEPTION: {info}")
    sys.__excepthook__(exctype, value, tb)


#-----------------------------------------------------------------------------------------------------------------------
def main():
    sys.excepthook = exception_logger
    os.environ['QT_MAC_WANTS_LAYER'] = '1'    # Workaround for https://bugreports.qt.io/browse/QTBUG-87014

    own_path = os.path.dirname(os.path.realpath(__file__)) + os.sep
    db, error = init_and_check_db(own_path)

    if error.code == LedgerInitError.EmptyDbInitialized:
        db, error = init_and_check_db(own_path)

    app = QApplication([])
    language = get_language(db)
    translator = QTranslator(app)
    language_file = own_path + "languages" + os.sep + language + '.qm'
    translator.load(language_file)
    app.installTranslator(translator)

    if db is None:
        window = AbortWindow(error.message)
    else:
        window = MainWindow(db, own_path, language)
    window.show()

    app.exec_()
    app.removeTranslator(translator)



import logging
from datetime import datetime

from PySide2.QtCore import Qt, Signal, Property, Slot, QEvent
from PySide2.QtSql import QSqlRelationalDelegate
from PySide2.QtWidgets import QDialog, QMessageBox
from PySide2.QtWidgets import QStyledItemDelegate

from jal.ui.ui_reference_data_dlg import Ui_ReferenceDataDialog
import jal.ui_custom.reference_selector as ui     # Full import due to "cyclic" reference
from jal.ui_custom.helpers import g_tr, UseSqlTable, ConfigureTableView, rel_idx
from jal.db.helpers import readSQL


# --------------------------------------------------------------------------------------------------------------
# Class to display and edit table with reference data (accounts, categories, tags...)
# --------------------------------------------------------------------------------------------------------------
class ReferenceDataDialog(QDialog, Ui_ReferenceDataDialog):
    # ----------------------------------------------------------------------------------------------------------
    # Params:
    # db - QSqlDatabase object for DB operations
    # table - name of the table to display/edit
    # columns - list of tuples - see helpers.py for details
    # title - title of dialog window
    # search_field - field name which will be used for search from GUI
    # tree_view - table will be displayed as hierarchical tree with help of 3 columns: 'id', 'pid' and 'children_count'
    #  ('pid' will identify parent row for current row, and '+' will be displayed for row with 'children_count'>0
    # relations - list of tuples that define lookup relations to other tables in database:
    def __init__(self, db, table, columns, title='',
                 search_field=None, toggle=None, tree_view=False, relations=None):
        QDialog.__init__(self)
        self.setupUi(self)

        self.selected_id = 0
        self.p_selected_name = ''
        self.dialog_visible = False
        self.selection_enabled = False
        self.tree_view = tree_view
        self.parent = 0
        self.last_parent = 0
        self.group_id = None
        self.group_key_field = None
        self.group_key_index = None
        self.group_fkey_field = None
        self.toggle_state = False
        self.toggle_field = None
        self.search_text = ""
        self.search_field = search_field

        self.db = db
        self.table = table
        self.Model = UseSqlTable(self, self.table, columns, relations)
        self.delegates = ConfigureTableView(self.DataView, self.Model, columns)
        # Storage of delegates inside class is required to keep ownership and prevent SIGSEGV as
        # https://doc.qt.io/qt-5/qabstractitemview.html#setItemDelegateForColumn says:
        # Any existing column delegate for column will be removed, but not deleted.
        # QAbstractItemView does not take ownership of delegate.

        self.GroupLbl.setVisible(False)
        self.GroupCombo.setVisible(False)
        if relations is not None:
            for relation in relations:
                if relation[rel_idx.GROUP_NAME] is not None:
                    self.GroupLbl.setVisible(True)
                    self.GroupLbl.setText(relation[rel_idx.GROUP_NAME])
                    self.GroupCombo.setVisible(True)
                    self.group_key_field = relation[rel_idx.KEY_FIELD]
                    self.group_key_index = self.Model.fieldIndex(relation[rel_idx.KEY_FIELD])
                    self.group_fkey_field = relation[rel_idx.FOREIGN_KEY]
                    relation_model = self.Model.relationModel(self.group_key_index)
                    self.GroupCombo.setModel(relation_model)
                    self.GroupCombo.setModelColumn(relation_model.fieldIndex(relation[rel_idx.LOOKUP_FIELD]))
                    self.group_id = relation_model.data(relation_model.index(0,
                                    relation_model.fieldIndex(self.group_fkey_field)))

        self.Toggle.setVisible(False)
        if toggle:
            self.Toggle.setVisible(True)
            self.toggle_field = toggle[0]
            self.Toggle.setText(toggle[1])

        self.setWindowTitle(title)
        if self.search_field is not None:
            self.SearchFrame.setVisible(True)
        else:
            self.SearchFrame.setVisible(False)
        self.UpBtn.setVisible(self.tree_view)

        self.SearchString.textChanged.connect(self.OnSearchChange)
        self.UpBtn.clicked.connect(self.OnUpClick)
        self.GroupCombo.currentIndexChanged.connect(self.OnGroupChange)
        self.Toggle.stateChanged.connect(self.OnToggleChange)
        self.AddBtn.clicked.connect(self.OnAdd)
        self.RemoveBtn.clicked.connect(self.OnRemove)
        self.CommitBtn.clicked.connect(self.OnCommit)
        self.RevertBtn.clicked.connect(self.OnRevert)
        self.DataView.clicked.connect(self.OnClicked)
        self.DataView.doubleClicked.connect(self.OnDoubleClicked)
        self.DataView.selectionModel().selectionChanged.connect(self.OnRowSelected)
        self.Model.dataChanged.connect(self.OnDataChanged)

        self.Model.select()
        self.setFilter()

    @Slot()
    def closeEvent(self, event):
        if self.CommitBtn.isEnabled():    # There are uncommited changed in a table
            if QMessageBox().warning(None, g_tr('ReferenceDataDialog', "Confirmation"),
                                     g_tr('ReferenceDataDialog', "You have uncommited changes. Do you want to close?"),
                                     QMessageBox.Yes, QMessageBox.No) == QMessageBox.No:
                event.ignore()
                return
            else:
                self.Model.revertAll()
        event.accept()

    # Overload ancestor method to activate/deactivate filters for table view
    def exec_(self, enable_selection=False):
        self.dialog_visible = True
        self.selection_enabled = enable_selection
        self.setFilter()
        res = super().exec_()
        self.dialog_visible = False
        self.resetFilter()
        return res

    def getSelectedName(self):
        if self.selected_id == 0:
            return g_tr('ReferenceDataDialog', "ANY")
        else:
            return self.p_selected_name

    def setSelectedName(self, selected_id):
        pass

    @Signal
    def selected_name_changed(self):
        pass

    SelectedName = Property(str, getSelectedName, setSelectedName, notify=selected_name_changed)

    @Slot()
    def OnDataChanged(self):
        self.CommitBtn.setEnabled(True)
        self.RevertBtn.setEnabled(True)

    @Slot()
    def OnAdd(self):
        new_record = self.Model.record()
        if self.tree_view:
            new_record.setValue('pid', self.parent)  # set current parent
        assert self.Model.insertRows(0, 1)
        self.Model.setRecord(0, new_record)
        self.CommitBtn.setEnabled(True)
        self.RevertBtn.setEnabled(True)

    @Slot()
    def OnRemove(self):
        idx = self.DataView.selectionModel().selection().indexes()
        selected_row = idx[0].row()
        assert self.Model.removeRow(selected_row)
        self.CommitBtn.setEnabled(True)
        self.RevertBtn.setEnabled(True)

    @Slot()
    def OnCommit(self):
        if self.group_key_index is not None:
            record = self.Model.record(0)
            group_field = record.value(self.Model.fieldIndex(self.group_key_field))
            if not group_field:
                self.Model.setData(self.Model.index(0, self.group_key_index), self.group_id)
        if not self.Model.submitAll():
            logging.fatal(g_tr('ReferenceDataDialog', "Submit failed: ") + self.Model.lastError().text())
            return
        self.CommitBtn.setEnabled(False)
        self.RevertBtn.setEnabled(False)

    @Slot()
    def OnRevert(self):
        self.Model.revertAll()
        self.CommitBtn.setEnabled(False)
        self.RevertBtn.setEnabled(False)

    def resetFilter(self):
        self.DataView.model().setFilter("")

    def setFilter(self):  # TODO: correctly combine different conditions
        if not self.dialog_visible:
            return

        conditions = []
        if self.search_text:
            conditions.append(f"{self.search_field} LIKE '%{self.search_text}%'")
        else:
            if self.tree_view:
                conditions.append(f"pid={self.parent}")

        if self.group_id:
            conditions.append(f"{self.table}.{self.group_key_field}={self.group_id}")

        if self.toggle_field:
            if not self.toggle_state:
                conditions.append(f"{self.table}.{self.toggle_field}=1")

        condition = ""
        for line in conditions:
            condition += line + " AND "
        condition = condition[:-len(" AND ")]

        self.DataView.model().setFilter(condition)

    @Slot()
    def OnSearchChange(self):
        self.search_text = self.SearchString.text()
        self.setFilter()

    @Slot()
    def OnRowSelected(self, selected, _deselected):
        idx = selected.indexes()
        if idx:
            selected_row = idx[0].row()
            self.selected_id = self.DataView.model().record(selected_row).value('id')
            self.p_selected_name = self.DataView.model().record(selected_row).value('name')

    @Slot()
    def OnClicked(self, index):
        if index.column() == 0:
            selected_row = index.row()
            self.parent = self.DataView.model().record(selected_row).value('id')
            self.last_parent = self.DataView.model().record(selected_row).value('pid')
            if self.search_text:
                self.SearchString.setText('')  # it will also call self.setFilter()
            else:
                self.setFilter()

    @Slot()
    def OnDoubleClicked(self, index):
        self.selected_id = self.DataView.model().record(index.row()).value('id')
        self.p_selected_name = self.DataView.model().record(index.row()).value('name')
        if self.selection_enabled:
            self.setResult(QDialog.Accepted)
            self.close()

    @Slot()
    def OnUpClick(self):
        if self.search_text:  # list filtered by search string
            return
        current_id = self.DataView.model().record(0).value('id')
        if current_id is None:
            pid = self.last_parent
        else:
            pid = readSQL(self.db,
                          f"SELECT c2.pid FROM {self.table} AS c1 LEFT JOIN {self.table} AS c2 ON c1.pid=c2.id "\
                          f"WHERE c1.id = :current_id", [(":current_id", current_id)])
            if pid == '':
                pid = 0
        self.parent = pid
        self.setFilter()

    @Slot()
    def OnGroupChange(self, list_id):
        model = self.GroupCombo.model()
        self.group_id = model.data(model.index(list_id, model.fieldIndex(self.group_fkey_field)))
        self.setFilter()

    @Slot()
    def OnToggleChange(self, state):
        if state == 0:
            self.toggle_state = False
        else:
            self.toggle_state = True
        self.setFilter()

# ===================================================================================================================
# Delegates to customize view of columns
# ===================================================================================================================

# -------------------------------------------------------------------------------------------------------------------
# Display '+' if element have children
class ReferenceTreeDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        children_count = model.data(model.index(index.row(), model.fieldIndex('children_count')), Qt.DisplayRole)
        text = ''
        if children_count:
            text = '+'
        painter.drawText(option.rect, Qt.AlignHCenter, text)
        painter.restore()

# -------------------------------------------------------------------------------------------------------------------
# Display '*' if true and empty cell if false
# Toggle True/False by mouse click
class ReferenceBoolDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        status = model.data(index, Qt.DisplayRole)
        if status:
            text = ' * '
        else:
            text = ''
        painter.drawText(option.rect, Qt.AlignHCenter, text)
        painter.restore()

    def editorEvent(self, event, model, option, index):
        if event.type() == QEvent.MouseButtonPress:
            if model.data(index, Qt.DisplayRole):  # Toggle value - from 1 to 0 and from 0 to 1
                model.setData(index, 0)
            else:
                model.setData(index, 1)
        return True

# -------------------------------------------------------------------------------------------------------------------
# Make integer alignment to the right
class ReferenceIntDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        value = model.data(index, Qt.DisplayRole)
        painter.drawText(option.rect, Qt.AlignRight, f"{value} ")
        painter.restore()

# -------------------------------------------------------------------------------------------------------------------
# Format unix timestamp into readable form '%d/%m/%Y %H:%M:%S'
class ReferenceTimestampDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        painter.save()
        model = index.model()
        timestamp = model.data(index, Qt.DisplayRole)
        if timestamp:
            text = datetime.fromtimestamp(timestamp).strftime('%d/%m/%Y %H:%M:%S')
        else:
            text = ""
        painter.drawText(option.rect, Qt.AlignLeft, text)
        painter.restore()

# -------------------------------------------------------------------------------------------------------------------
# The class itself is empty but it activates built-in editors for lookup tables
class ReferenceLookupDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

# -----------------------------------------------------------------------------------------------------------------------
# Delegate to display tag editor
class ReferencePeerDelegate(QSqlRelationalDelegate):
    def __init__(self, parent=None):
        QSqlRelationalDelegate.__init__(self, parent)

    def createEditor(self, aParent, option, index):
        peer_selector = ui.PeerSelector(aParent)
        peer_selector.init_db(index.model().database())
        return peer_selector

    def setModelData(self, editor, model, index):
        model.setData(index, editor.selected_id)


import datetime

from jal.constants import ColumnWidth
from PySide2.QtCore import QCoreApplication, Qt, QStringListModel, QByteArray
from PySide2.QtSql import QSqlTableModel, QSqlRelationalTableModel, QSqlRelation
from PySide2.QtWidgets import QHeaderView, QDataWidgetMapper, QFrame, QDateTimeEdit

class hcol_idx:
    DB_NAME = 0
    DISPLAY_NAME = 1
    WIDTH = 2
    SORT_ORDER = 3
    DELEGATE = 4

class rel_idx:
    KEY_FIELD = 0
    LOOKUP_TABLE = 1
    FOREIGN_KEY = 2
    LOOKUP_FIELD = 3
    GROUP_NAME = 4          # Name of group field if Reference Data dialog

class map_idx:
    DB_NAME = 0
    WIDGET = 1


# -----------------------------------------------------------------------------------------------------------------------
# Global translate helper to make lines shorter in code
def g_tr(context, text):
    return QCoreApplication.translate(context, text)


# -----------------------------------------------------------------------------------------------------------------------
# Returns True if all modules from module_list are present in the system
def dependency_present(module_list):
    result = True
    for module in module_list:
        try:
            __import__(module)
        except ImportError:
            result = False
    return result


# -----------------------------------------------------------------------------------------------------------------------
# Helpers to work with numbers

def formatFloatLong(value):
    if abs(value - round(value)) <= 10e-2:
        text = f"{value:.0f}"
    elif abs(value - round(value, 2)) <= 10e-4:
        text = f"{value:.2f}"
    elif abs(value - round(value, 4)) <= 10e-6:
        text = f"{value:.4f}"
    elif abs(value - round(value, 6)) <= 10e-8:
        text = f"{value:.6f}"
    else:
        text = f"{value:.8f}"
    return text


# -----------------------------------------------------------------------------------------------------------------------
# a simple VLine, like the one from Qt Designer to use in Status Bar for example
class VLine(QFrame):
    def __init__(self):
        super(VLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)


# -----------------------------------------------------------------------------------------------------------------------
# Helpers to work with datetime
class ManipulateDate:
    @staticmethod
    def toTimestamp(date_value):
        time_value = datetime.time(0, 0, 0)
        dt_value = datetime.datetime.combine(date_value, time_value)
        return int(dt_value.timestamp())

    @staticmethod
    def startOfPreviousWeek():
        prev_week = datetime.date.today() - datetime.timedelta(days = 7)
        start_of_week = prev_week - datetime.timedelta(days = prev_week.weekday())
        return ManipulateDate.toTimestamp(start_of_week)

    @staticmethod
    def startOfPreviousMonth():
        day = datetime.date.today()
        first_day_of_month = day.replace(day=1)
        last_day_of_prev_month = first_day_of_month - datetime.timedelta(days=1)
        first_day_of_prev_month = last_day_of_prev_month.replace(day=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_month)

    @staticmethod
    def startOfPreviousQuarter():
        day = datetime.date.today()
        prev_quarter_month = day.month - day.month % 3 - 3
        if prev_quarter_month > 0:
            quarter_back = day.replace(month = prev_quarter_month)
        else:
            quarter_back = day.replace(month = (prev_quarter_month + 12), year = (day.year - 1))
        first_day_of_prev_quarter = quarter_back.replace(day=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_quarter)

    @staticmethod
    def startOfPreviousYear():
        day = datetime.date.today()
        first_day_of_year = day.replace(day=1, month=1)
        last_day_of_prev_year = first_day_of_year - datetime.timedelta(days=1)
        first_day_of_prev_year = last_day_of_prev_year.replace(day=1, month=1)
        return ManipulateDate.toTimestamp(first_day_of_prev_year)

    @staticmethod
    def Last3Months():
        day = datetime.date.today()
        end = day + datetime.timedelta(days=1)
        begin_month = day.month - 3
        if begin_month > 0:
            begin = day.replace(month=begin_month)
        else:
            begin = day.replace(month=(begin_month + 12), year=(day.year - 1))
        begin = begin.replace(day=1)
        return (ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end))

    @staticmethod
    def RangeYTD():
        day = datetime.date.today()
        end = day + datetime.timedelta(days=1)
        begin = day.replace(day=1, year=(day.year - 1))
        return (ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end))

    @staticmethod
    def RangeThisYear():
        day = datetime.date.today()
        end = day + datetime.timedelta(days=1)
        begin = day.replace(day=1, month=1)
        return (ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end))

    @staticmethod
    def RangePreviousYear():
        day = datetime.date.today()
        end = day.replace(day=1, month=1)
        begin = end.replace(year=(day.year - 1))
        return (ManipulateDate.toTimestamp(begin), ManipulateDate.toTimestamp(end))


# -------------------------------------------------------------------------------------------------------------------
# column_list is a list of tuples: (db_column_name, display_column_name, width, sort_order, delegate)
# column will be hidden if display_column_name is None
# column with negative with will be stretched
# sort order should be from Qt module (eg. Qt.AscendingOrder) or None
# delegate is a function for custom paint and editors
# relations - list of tuples that define lookup relations to other tables in database:
#             [(KEY_FEILD, LOOKUP_TABLE, FOREIGN_KEY, LOOKUP_FIELD), ...]
# Returns QSqlTableModel/QSqlRelationalQueryModel
def UseSqlTable(parent, table_name, columns, relations):
    if relations:
        model = QSqlRelationalTableModel(parent=parent, db=parent.db)
    else:
        model = QSqlTableModel(parent=parent, db=parent.db)
    model.setTable(table_name)
    model.setEditStrategy(QSqlTableModel.OnManualSubmit)
    if relations:
        model.setJoinMode(QSqlRelationalTableModel.LeftJoin)  # to work correctly with NULL values in fields
        for relation in relations:
            model.setRelation(model.fieldIndex(relation[rel_idx.KEY_FIELD]),
                              QSqlRelation(relation[rel_idx.LOOKUP_TABLE],
                                           relation[rel_idx.FOREIGN_KEY], relation[rel_idx.LOOKUP_FIELD]))
    for column in columns:
        if column[hcol_idx.DISPLAY_NAME]:
            model.setHeaderData(model.fieldIndex(column[hcol_idx.DB_NAME]), Qt.Horizontal,
                                g_tr('TableViewConfig', column[hcol_idx.DISPLAY_NAME]))
        if column[hcol_idx.SORT_ORDER] is not None:
            model.setSort(model.fieldIndex(column[hcol_idx.DB_NAME]), column[hcol_idx.SORT_ORDER])
    return model

# -------------------------------------------------------------------------------------------------------------------
# column_list is a list of tuples: (db_column_name, display_column_name, width, sort_order, delegate)
# column will be hidden if display_column_name is None
# column with negative with will be stretched
# sort order is ignored as it might be set by Query itself
# delegate is a function for custom paint and editors
# Returns : QSqlTableModel
def UseSqlQuery(parent, query, columns):
    model = QSqlTableModel(parent=parent, db=parent.db)
    model.setQuery(query)
    for column in columns:
        if column[hcol_idx.DISPLAY_NAME]:
            model.setHeaderData(model.fieldIndex(column[hcol_idx.DB_NAME]), Qt.Horizontal, column[hcol_idx.DISPLAY_NAME])
    return model

# -------------------------------------------------------------------------------------------------------------------
# Use mappings to link between DB fields and GUI widgets with help of delegate
# mapping is a list of tuples [(FIELD_NAME, GUI_WIDGET)]
# If widget is a custom one:
#    - initialize database connection for it
#    - connect "changed" signal to "submit" slot of QDataWidgetMapper (to reflect data changes in UI)
def ConfigureDataMappers(model, mappings, delegate):
    mapper = QDataWidgetMapper(model)
    mapper.setModel(model)
    mapper.setSubmitPolicy(QDataWidgetMapper.AutoSubmit)
    mapper.setItemDelegate(delegate(mapper))
    for mapping in mappings:
        if hasattr(mapping[map_idx.WIDGET], "isCustom"):
            mapping[map_idx.WIDGET].init_db(model.database())
            mapping[map_idx.WIDGET].changed.connect(mapper.submit)
        if mapping[map_idx.WIDGET].property("stringModelData") is not None:
            # Load lookup values from dynamic property into combo-box model
            comboModel = QStringListModel(mapping[map_idx.WIDGET].property("stringModelData").split(';'))
            mapping[map_idx.WIDGET].setModel(comboModel)
            mapper.addMapping(mapping[map_idx.WIDGET], model.fieldIndex(mapping[map_idx.DB_NAME]),
                              QByteArray().setRawData("currentIndex", 12))
        else:  # ComboBox has no USER property so we use QByteArray above, and simple mapping for others below
            mapper.addMapping(mapping[map_idx.WIDGET], model.fieldIndex(mapping[map_idx.DB_NAME]))
        # adjust width of QDateTimeEdits to show full date-time string
        if isinstance(mapping[map_idx.WIDGET], QDateTimeEdit):
            mapping[map_idx.WIDGET].setFixedWidth(
                mapping[map_idx.WIDGET].fontMetrics().width("00/00/0000 00:00:00") * 1.25)
    return mapper

# -------------------------------------------------------------------------------------------------------------------
# Return value is a list of delegates because storage of delegates
# is required to keep ownership and prevent SIGSEGV as
# https://doc.qt.io/qt-5/qabstractitemview.html#setItemDelegateForColumn says:
# Any existing column delegate for column will be removed, but not deleted.
# QAbstractItemView does not take ownership of delegate.
def ConfigureTableView(view, model, columns):
    view.setModel(model)
    for column in columns:
        if column[hcol_idx.DISPLAY_NAME] is None:   # hide column
            view.setColumnHidden(model.fieldIndex(column[hcol_idx.DB_NAME]), True)
        if column[hcol_idx.WIDTH] is not None:
            if column[hcol_idx.WIDTH] == ColumnWidth.STRETCH:
                view.horizontalHeader().setSectionResizeMode(model.fieldIndex(column[hcol_idx.DB_NAME]),
                                                             QHeaderView.Stretch)
            elif column[hcol_idx.WIDTH] == ColumnWidth.FOR_DATETIME:
                view.setColumnWidth(model.fieldIndex(column[hcol_idx.DB_NAME]),
                                    view.fontMetrics().width("00/00/0000 00:00:00") * 1.1)
            else:  # set custom width
                view.setColumnWidth(model.fieldIndex(column[hcol_idx.DB_NAME]), column[hcol_idx.WIDTH])

    font = view.horizontalHeader().font()
    font.setBold(True)
    view.horizontalHeader().setFont(font)

    delegates = []
    for column in columns:
        if column[hcol_idx.DELEGATE] is None:
            # Use standard delegate / Remove old delegate if there was any
            view.setItemDelegateForColumn(model.fieldIndex(column[hcol_idx.DB_NAME]), None)
        else:
            delegates.append(column[hcol_idx.DELEGATE](view))
            view.setItemDelegateForColumn(model.fieldIndex(column[hcol_idx.DB_NAME]), delegates[-1])
    return delegates




class xslxFormat:
    def __init__(self, workbook):
        self.wbk = workbook
        self.even_color_bg = '#C0C0C0'
        self.odd_color_bg = '#FFFFFF'

    def Bold(self):
        return self.wbk.add_format({'bold': True})

    def ColumnHeader(self):
        return self.wbk.add_format({'bold': True,
                                    'text_wrap': True,
                                    'align': 'center',
                                    'valign': 'vcenter',
                                    'bg_color': '#808080',
                                    'font_color': '#FFFFFF',
                                    'border': 1})

    def ColumnFooter(self):
        return self.wbk.add_format({'bold': True,
                                    'num_format': '#,###,##0.00',
                                    'bg_color': '#808080',
                                    'font_color': '#FFFFFF',
                                    'border': 1})

    def Text(self, even_odd_value=1):
        if even_odd_value % 2:
            bg_color = self.odd_color_bg
        else:
            bg_color = self.even_color_bg
        return self.wbk.add_format({'border': 1,
                                    'valign': 'vcenter',
                                    'bg_color': bg_color})

    def Number(self, even_odd_value=1, tolerance=2, center=False):
        if even_odd_value % 2:
            bg_color = self.odd_color_bg
        else:
            bg_color = self.even_color_bg
        num_format = ''
        if tolerance > 0:
            num_format = '#,###,##0.'
            for i in range(tolerance):
                num_format = num_format + '0'
        if center:
            align = 'center'
        else:
            align = 'right'
        return self.wbk.add_format({'num_format': num_format,
                                    'border': 1,
                                    'align': align,
                                    'valign': 'vcenter',
                                    'bg_color': bg_color})

#-----------------------------------------------------------------------------------------------------------------------
ROW_DATA = 0
ROW_FORMAT = 1
ROW_WIDTH = 2
ROW_SPAN_H = 3
ROW_SPAN_V = 4


def xlsxWriteRow(wksheet, row, columns, height=None):
    if height:
        wksheet.set_row(row, height)
    for column in columns:
        cd = columns[column]
        if len(cd) != 2:
            if cd[ROW_WIDTH]:
                wksheet.set_column(column, column, cd[ROW_WIDTH])
            if cd[ROW_SPAN_H] or cd[ROW_SPAN_V]:
                wksheet.merge_range(row, column, row+cd[ROW_SPAN_V], column+cd[ROW_SPAN_H],
                                    cd[ROW_DATA], cd[ROW_FORMAT])
        wksheet.write(row, column, cd[ROW_DATA], cd[ROW_FORMAT])


import logging

from datetime import datetime
from jal.constants import Setup, BookAccount, TransactionType, ActionSubtype, TransferSubtype, CorporateAction, \
    PredefinedCategory, PredefinedPeer
from PySide2.QtCore import Qt, QDate, QDateTime
from PySide2.QtWidgets import QDialog, QMessageBox
from jal.db.helpers import executeSQL, readSQL, readSQLrecord
from jal.db.routines import calculateBalances, calculateHoldings
from jal.ui_custom.helpers import g_tr
from jal.ui.ui_rebuild_window import Ui_ReBuildDialog


class RebuildDialog(QDialog, Ui_ReBuildDialog):
    def __init__(self, parent, frontier):
        QDialog.__init__(self)
        self.setupUi(self)

        self.LastRadioButton.toggle()
        self.frontier = frontier
        frontier_text = datetime.fromtimestamp(frontier).strftime('%d/%m/%Y')
        self.FrontierDateLabel.setText(frontier_text)
        self.CustomDateEdit.setDate(QDate.currentDate())

        # center dialog with respect to parent window
        x = parent.x() + parent.width()/2 - self.width()/2
        y = parent.y() + parent.height()/2 - self.height()/2
        self.setGeometry(x, y, self.width(), self.height())

    def isFastAndDirty(self):
        return self.FastAndDirty.isChecked()

    def getTimestamp(self):
        if self.LastRadioButton.isChecked():
            return self.frontier
        elif self.DateRadionButton.isChecked():
            return self.CustomDateEdit.dateTime().toSecsSinceEpoch()
        else:  # self.AllRadioButton.isChecked()
            return 0


# ===================================================================================================================
# TODO Check are there positive lines for Incomes
# TODO Check are there negative lines for Costs
# ===================================================================================================================
class Ledger:
    SILENT_REBUILD_THRESHOLD = 50

    def __init__(self, db):
        self.db = db
        self.current = {}
        self.current_seq = -1
        self.balances_view = None
        self.holdings_view = None
        self.balance_active_only = 1
        self.balance_currency = None
        self.balance_date = QDateTime.currentSecsSinceEpoch()
        self.holdings_date = QDateTime.currentSecsSinceEpoch()
        self.holdings_currency = None

    def setViews(self, balances, holdings):
        self.balances_view = balances
        self.holdings_view = holdings

    def setActiveBalancesOnly(self, active_only):
        if self.balance_active_only != active_only:
            self.balance_active_only = active_only
            self.updateBalancesView()

    def setBalancesDate(self, balance_date):
        if self.balance_date != balance_date:
            self.balance_date = balance_date
            self.updateBalancesView()

    def setBalancesCurrency(self, currency_id, currency_name):
        if self.balance_currency != currency_id:
            self.balance_currency = currency_id
            balances_model = self.balances_view.model()
            balances_model.setHeaderData(balances_model.fieldIndex("balance_adj"), Qt.Horizontal,
                                         g_tr('Ledger', "Balance, ") + currency_name)
            self.updateBalancesView()

    def updateBalancesView(self):
        calculateBalances(self.db, self.balance_date, self.balance_currency, self.balance_active_only)
        self.balances_view.model().select()

    def setHoldingsDate(self, holdings_date):
        if self.holdings_date != holdings_date:
            self.holdings_date = holdings_date
            self.updateHoldingsView()

    def setHoldingsCurrency(self, currency_id, currency_name):
        if self.holdings_currency != currency_id:
            self.holdings_currency = currency_id
            holidings_model = self.holdings_view.model()
            holidings_model.setHeaderData(holidings_model.fieldIndex("value_adj"), Qt.Horizontal,
                                          g_tr('Ledger', "Value, ") + currency_name)
            self.updateHoldingsView()

    def updateHoldingsView(self):
        calculateHoldings(self.db, self.holdings_date, self.holdings_currency)
        holdings_model = self.holdings_view.model()
        holdings_model.select()
        for row in range(holdings_model.rowCount()):
            if holdings_model.data(holdings_model.index(row, 1)):
                self.holdings_view.setSpan(row, 3, 1, 3)
        self.holdings_view.show()

    # Returns timestamp of last operations that were calculated into ledger
    def getCurrentFrontier(self):
        current_frontier = readSQL(self.db, "SELECT ledger_frontier FROM frontier")
        if current_frontier == '':
            current_frontier = 0
        return current_frontier

    # Add one more transaction to 'book' of ledger.
    # If book is Assets and value is not None then amount contains Asset Quantity and Value contains amount
    #    of money in current account currency. Otherwise Amount contains only money value.
    # Method uses Account, Asset,Peer, Category and Tag values from current transaction
    def appendTransaction(self, book, amount, value=None):
        seq_id = self.current_seq
        timestamp = self.current['timestamp']
        if book == BookAccount.Assets:
            asset_id = self.current['asset']
        else:
            asset_id = self.current['currency']
        account_id = self.current['account']
        if book == BookAccount.Costs or book == BookAccount.Incomes:
            peer_id = self.current['peer']
            category_id = self.current['category']
            tag_id = None if self.current['tag'] == '' else self.current['tag']
        else:
            peer_id = None
            category_id = None
            tag_id = None
        try:
            old_sid, old_amount, old_value = readSQL(self.db,
                "SELECT sid, sum_amount, sum_value FROM ledger_sums "
                "WHERE book_account = :book AND asset_id = :asset_id "
                "AND account_id = :account_id AND sid <= :seq_id "
                "ORDER BY sid DESC LIMIT 1",
                [(":book", book), (":asset_id", asset_id), (":account_id", account_id), (":seq_id", seq_id)])
        except:
            old_sid = -1
            old_amount = 0.0
            old_value = 0.0
        new_amount = old_amount + amount
        if value is None:
            new_value = old_value
        else:
            new_value = old_value + value
        if (abs(new_amount - old_amount) + abs(new_value - old_value)) <= (2 * Setup.CALC_TOLERANCE):
            return  # we have zero amount - no reason to put it into ledger

        _ = executeSQL(self.db, "INSERT INTO ledger (timestamp, sid, book_account, asset_id, account_id, "
                                "amount, value, peer_id, category_id, tag_id) "
                                "VALUES(:timestamp, :sid, :book, :asset_id, :account_id, "
                                ":amount, :value, :peer_id, :category_id, :tag_id)",
                       [(":timestamp", timestamp), (":sid", seq_id), (":book", book), (":asset_id", asset_id),
                        (":account_id", account_id), (":amount", amount), (":value", value),
                        (":peer_id", peer_id), (":category_id", category_id), (":tag_id", tag_id)])
        if seq_id == old_sid:
            _ = executeSQL(self.db, "UPDATE ledger_sums SET sum_amount = :new_amount, sum_value = :new_value"
                                    " WHERE sid = :sid AND book_account = :book"
                                    " AND asset_id = :asset_id AND account_id = :account_id",
                           [(":new_amount", new_amount), (":new_value", new_value), (":sid", seq_id),
                            (":book", book), (":asset_id", asset_id), (":account_id", account_id)])
        else:
            _ = executeSQL(self.db, "INSERT INTO ledger_sums(sid, timestamp, book_account, "
                                    "asset_id, account_id, sum_amount, sum_value) "
                                    "VALUES(:sid, :timestamp, :book, :asset_id, "
                                    ":account_id, :new_amount, :new_value)",
                           [(":sid", seq_id), (":timestamp", timestamp), (":book", book), (":asset_id", asset_id),
                            (":account_id", account_id), (":new_amount", new_amount), (":new_value", new_value)])
        self.db.commit()

    # TODO check that condition <= is really correct for timestamp in this function
    # Returns Amount measured in current account currency or asset_id that 'book' has at current ledger frontier
    def getAmount(self, book, asset_id=None):
        if asset_id is None:
            amount = readSQL(self.db,
                               "SELECT sum_amount FROM ledger_sums WHERE book_account = :book AND "
                               "account_id = :account_id AND timestamp <= :timestamp ORDER BY sid DESC LIMIT 1",
                               [(":book", book), (":account_id", self.current['account']),
                                (":timestamp", self.current['timestamp'])])
        else:
            amount = readSQL(self.db, "SELECT sum_amount FROM ledger_sums WHERE book_account = :book "
                                        "AND account_id = :account_id AND asset_id = :asset_id "
                                        "AND timestamp <= :timestamp ORDER BY sid DESC LIMIT 1",
                               [(":book", book), (":account_id", self.current['account']),
                                (":asset_id", asset_id), (":timestamp", self.current['timestamp'])])
        amount = float(amount) if amount is not None else 0.0
        return amount

    def takeCredit(self, operation_amount):
        money_available = self.getAmount(BookAccount.Money)
        credit = 0
        if money_available < operation_amount:
            credit = operation_amount - money_available
            self.appendTransaction(BookAccount.Liabilities, -credit)
        return credit

    def returnCredit(self, operation_amount):
        current_credit_value = -1.0 * self.getAmount(BookAccount.Liabilities)
        debit = 0
        if current_credit_value > 0:
            if current_credit_value >= operation_amount:
                debit = operation_amount
            else:
                debit = current_credit_value
        if debit > 0:
            self.appendTransaction(BookAccount.Liabilities, debit)
        return debit

    def processActionDetails(self):
        query = executeSQL(self.db, "SELECT sum as amount, category_id, tag_id FROM action_details AS d WHERE pid=:pid",
                           [(":pid", self.current['id'])])
        while query.next():
            amount, self.current['category'], self.current['tag'] = readSQLrecord(query)
            book = BookAccount.Costs if amount < 0 else BookAccount.Incomes
            self.appendTransaction(book, -amount)

    def processAction(self):
        action_amount = self.current['amount']
        if action_amount < 0:
            credit_taken = self.takeCredit(-action_amount)
            self.appendTransaction(BookAccount.Money, -(-action_amount - credit_taken))
        else:
            credit_returned = self.returnCredit(action_amount)
            if credit_returned < action_amount:
                self.appendTransaction(BookAccount.Money, action_amount - credit_returned)
        if self.current['subtype'] == ActionSubtype.SingleIncome:
            self.appendTransaction(BookAccount.Incomes, -action_amount)
        elif self.current['subtype'] == ActionSubtype.SingleSpending:
            self.appendTransaction(BookAccount.Costs, -action_amount)
        else:
            self.processActionDetails()

    def processDividend(self):
        if self.current['peer'] == '':
            logging.error(g_tr('Ledger', "Can't process dividend as bank isn't set for investment account"))
            return
        dividend_amount = self.current['amount']
        tax_amount = self.current['fee_tax']
        credit_returned = self.returnCredit(dividend_amount - tax_amount)
        if credit_returned < dividend_amount:
            self.appendTransaction(BookAccount.Money, dividend_amount - credit_returned)
        self.appendTransaction(BookAccount.Incomes, -dividend_amount)
        if tax_amount:
            self.appendTransaction(BookAccount.Money, -tax_amount)
            self.current['category'] = PredefinedCategory.Taxes
            self.appendTransaction(BookAccount.Costs, tax_amount)

    def processBuy(self):
        seq_id = self.current_seq
        account_id = self.current['account']
        asset_id = self.current['asset']
        qty = self.current['amount']
        price = self.current['price']
        trade_value = round(price * qty, 2) + self.current['fee_tax'] + self.current['coupon']
        sell_qty = 0
        sell_value = 0
        if self.getAmount(BookAccount.Assets, asset_id) < 0:        # Match deals if we have something sold before
            # Get information about last deal
            query = executeSQL(self.db,
                               "SELECT d.open_sid AS open, ABS(o.qty) - SUM(d.qty) AS remainder FROM deals AS d "
                               "LEFT JOIN sequence AS os ON os.type=3 AND os.id=d.open_sid "
                               "JOIN trades AS o ON o.id = os.operation_id "
                               "WHERE d.account_id=:account_id AND d.asset_id=:asset_id "
                               "GROUP BY d.open_sid "
                               "ORDER BY d.close_sid DESC, d.open_sid DESC LIMIT 1",
                               [(":account_id", account_id), (":asset_id", asset_id)])
            last_sid = 0  # Take all Sell operations by default
            reminder = 0
            if query.next():
                # Get sid of Sell trade from the last deal and non-matched reminder of last Sell trade
                last_sid, reminder = readSQLrecord(query)
                # if last Sell is fully matched (reminder>=0) we start from next trade, otherwise we need to shift by 1
                if reminder < 0:
                    last_sid = last_sid - 1
            query = executeSQL(self.db,
                               "SELECT s.id, -t.qty, t.price FROM trades AS t "
                               "LEFT JOIN sequence AS s ON s.type = 3 AND s.operation_id=t.id "
                               "WHERE t.qty < 0 AND t.asset_id = :asset_id AND t.account_id = :account_id "
                               "AND s.id < :sid AND s.id > :last_sid",
                               [(":asset_id", asset_id), (":account_id", account_id),
                                (":sid", seq_id), (":last_sid", last_sid)])
            while query.next():   # Perform match ("closure") of previous Sell trades
                deal_sid, deal_qty, deal_price = readSQLrecord(query)
                if reminder < 0:
                    next_deal_qty = reminder
                    reminder = 0
                else:
                    next_deal_qty = deal_qty
                if (sell_qty + next_deal_qty) >= qty:  # we are buying less or the same amount as was sold previously
                    next_deal_qty = qty - sell_qty
                _ = executeSQL(self.db, "INSERT INTO deals(account_id, asset_id, open_sid, close_sid, qty) "
                                        "VALUES(:account_id, :asset_id, :open_sid, :close_sid, :qty)",
                               [(":account_id", account_id), (":asset_id", asset_id), (":open_sid", deal_sid),
                                (":close_sid", seq_id), (":qty", -next_deal_qty)])
                sell_qty = sell_qty + next_deal_qty
                sell_value = sell_value + (next_deal_qty * deal_price)
                if sell_qty == qty:
                    break
        credit_taken = self.takeCredit(trade_value)
        if trade_value != credit_taken:
            self.appendTransaction(BookAccount.Money, -(trade_value - credit_taken))
        if sell_qty > 0:  # Add result of closed deals
            self.appendTransaction(BookAccount.Assets, sell_qty, sell_value)
            self.current['category'] = PredefinedCategory.Profit
            self.appendTransaction(BookAccount.Incomes, ((price * sell_qty) - sell_value))
        if sell_qty < qty:  # Add new long position
            self.appendTransaction(BookAccount.Assets, (qty - sell_qty), (qty - sell_qty) * price)
        if self.current['coupon']:
            self.current['category'] = PredefinedCategory.Dividends
            self.appendTransaction(BookAccount.Costs, self.current['coupon'])
        if self.current['fee_tax']:
            self.current['category'] = PredefinedCategory.Fees
            self.appendTransaction(BookAccount.Costs, self.current['fee_tax'])

    def processSell(self):
        seq_id = self.current_seq
        account_id = self.current['account']
        asset_id = self.current['asset']
        qty = -self.current['amount']
        price = self.current['price']
        trade_value = round(price * qty, 2) - self.current['fee_tax'] + self.current['coupon']
        buy_qty = 0
        buy_value = 0
        if self.getAmount(BookAccount.Assets, asset_id) > 0:    # Match deals if we have something bought before
            # Get information about last deal
            query = executeSQL(self.db,
                               "SELECT d.open_sid AS open, ABS(coalesce(o.qty, ca.qty_new))- SUM(d.qty) AS remainder "
                               "FROM deals AS d "
                               "LEFT JOIN sequence AS os ON os.id=d.open_sid "
                               "LEFT JOIN trades AS o ON o.id = os.operation_id AND os.type = 3 "
                               "LEFT JOIN corp_actions AS ca ON ca.id = os.operation_id AND os.type = 5 "
                               "WHERE d.account_id=:account_id AND d.asset_id=:asset_id "
                               "GROUP BY d.open_sid "
                               "ORDER BY d.close_sid DESC, d.open_sid DESC LIMIT 1",
                               [(":account_id", account_id), (":asset_id", asset_id)])
            last_sid = 0  # Take all Buy operations by default
            reminder = 0
            if query.next():     # Perform match ("closure") of previous Buy trades
                # Get sid of Buy trade from the last deal and non-matched reminder of last Buy trade
                last_sid, reminder = readSQLrecord(query)
                # if last Buy is fully matched (reminder<=0) we start from next trade, otherwise we need to shift by 1
                if reminder > 0:
                    last_sid = last_sid - 1
            query = executeSQL(self.db,
                               "SELECT * FROM ("
                               "SELECT s.id, t.qty, t.price FROM trades AS t "
                               "LEFT JOIN sequence AS s ON s.type = 3 AND s.operation_id=t.id "
                               "WHERE t.qty > 0 AND t.asset_id = :asset_id AND t.account_id = :account_id "
                               "AND s.id < :sid AND s.id > :last_sid "
                               "UNION ALL "
                               "SELECT s.id, c.qty_new, coalesce(l.value/c.qty_new, 0) AS price FROM corp_actions AS c "
                               "LEFT JOIN sequence AS s ON s.type = 5 AND s.operation_id=c.id " 
                               "LEFT JOIN ledger AS l ON s.id = l.sid AND l.asset_id=c.asset_id_new AND l.value > 0 "
                               "WHERE c.qty_new > 0 AND c.asset_id_new = :asset_id AND c.account_id = :account_id "
                               "AND s.id < :sid AND s.id > :last_sid " 
                               ")ORDER BY id",
                               [(":asset_id", asset_id), (":account_id", account_id),
                                (":sid", seq_id), (":last_sid", last_sid)])
            while query.next():    # Perform match ("closure") of previous Buy trades
                deal_sid, deal_qty, deal_price = readSQLrecord(query)  # deal_sid -> trade_sid
                if reminder > 0:
                    next_deal_qty = reminder
                    reminder = 0
                else:
                    next_deal_qty = deal_qty
                if (buy_qty + next_deal_qty) >= qty:  # we are selling less or the same amount as was bought previously
                    next_deal_qty = qty - buy_qty
                _ = executeSQL(self.db, "INSERT INTO deals(account_id, asset_id, open_sid, close_sid, qty) "
                                        "VALUES(:account_id, :asset_id, :open_sid, :close_sid, :qty)",
                               [(":account_id", account_id), (":asset_id", asset_id), (":open_sid", deal_sid),
                                (":close_sid", seq_id), (":qty", next_deal_qty)])
                buy_qty = buy_qty + next_deal_qty
                buy_value = buy_value + (next_deal_qty * deal_price)
                if buy_qty == qty:
                    break
        credit_returned = self.returnCredit(trade_value)
        if credit_returned < trade_value:
            self.appendTransaction(BookAccount.Money, (trade_value - credit_returned))
        if buy_qty > 0:  # Add result of closed deals
            self.appendTransaction(BookAccount.Assets, -buy_qty, -buy_value)
            self.current['category'] = PredefinedCategory.Profit
            self.appendTransaction(BookAccount.Incomes, (buy_value - (price * buy_qty)))
        if buy_qty < qty:  # Add new short position
            self.appendTransaction(BookAccount.Assets, (buy_qty - qty), (buy_qty - qty) * price)
        if self.current['coupon']:
            self.current['category'] = PredefinedCategory.Dividends
            self.appendTransaction(BookAccount.Incomes, -self.current['coupon'])
        if self.current['fee_tax']:
            self.current['category'] = PredefinedCategory.Fees
            self.appendTransaction(BookAccount.Costs, self.current['fee_tax'])

    def processTrade(self):
        if self.current['peer'] == '':
            logging.error(g_tr('Ledger', "Can't process trade as bank isn't set for investment account"))
            return
        operationTrade = {
            -1: self.processSell,
            1: self.processBuy
        }
        operationTrade[self.current['subtype']]()

    def processTransferOut(self):
        amount = -self.current['amount']
        credit_taken = self.takeCredit(amount)
        self.appendTransaction(BookAccount.Money, -(amount - credit_taken))
        self.appendTransaction(BookAccount.Transfers, amount)

    def processTransferIn(self):
        amount = self.current['amount']
        credit_returned = self.returnCredit(amount)
        if credit_returned < amount:
            self.appendTransaction(BookAccount.Money, (amount - credit_returned))
        self.appendTransaction(BookAccount.Transfers, -amount)

    def processTransferFee(self):
        fee = -self.current['amount']
        credit_taken = self.takeCredit(fee)
        self.current['peer'] = PredefinedPeer.Financial
        self.appendTransaction(BookAccount.Money, -(fee - credit_taken))
        self.current['category'] = PredefinedCategory.Fees
        self.appendTransaction(BookAccount.Costs, fee)

    def processTransfer(self):
        operationTransfer = {
            TransferSubtype.Outgoing: self.processTransferOut,
            TransferSubtype.Fee: self.processTransferFee,
            TransferSubtype.Incoming: self.processTransferIn
        }
        operationTransfer[self.current['subtype']]()

    def processAssetConversion(self):
        seq_id = self.current_seq
        account_id = self.current['account']
        asset_id = self.current['asset']
        qty = self.current['amount']
        buy_qty = 0
        buy_value = 0
        if self.getAmount(BookAccount.Assets, asset_id) < (qty - 2*Setup.CALC_TOLERANCE):
            logging.fatal(g_tr('Ledger', "Conversion failed. Asset amount is not enogh. Date: ")
                          + f"{datetime.fromtimestamp(self.current['timestamp']).strftime('%d/%m/%Y %H:%M:%S')}")
            return
        query = executeSQL(self.db,
                           "SELECT d.open_sid AS open, ABS(coalesce(o.qty, ca.qty_new))- SUM(d.qty) AS remainder "
                           "FROM deals AS d "
                           "LEFT JOIN sequence AS os ON os.id=d.open_sid "
                           "LEFT JOIN trades AS o ON o.id = os.operation_id AND os.type = 3 "
                           "LEFT JOIN corp_actions AS ca ON ca.id = os.operation_id AND os.type = 5 "
                           "WHERE d.account_id=:account_id AND d.asset_id=:asset_id "
                           "GROUP BY d.open_sid "
                           "ORDER BY d.close_sid DESC, d.open_sid DESC LIMIT 1",
                           [(":account_id", account_id), (":asset_id", asset_id)])
        last_sid = 0  # Take all Buy operations by default
        reminder = 0
        if query.next():  # Perform match ("closure") of previous Buy trades
            # Get sid of Buy trade from the last deal and non-matched reminder of last Buy trade
            last_sid, reminder = readSQLrecord(query)
            # if last Buy is fully matched (reminder<=0) we start from next trade, otherwise we need to shift by 1
            if reminder > 0:
                last_sid = last_sid - 1
        query = executeSQL(self.db,
                           "SELECT * FROM ("
                           "SELECT s.id, t.qty, t.price FROM trades AS t "
                           "LEFT JOIN sequence AS s ON s.type = 3 AND s.operation_id=t.id "
                           "WHERE t.qty > 0 AND t.asset_id = :asset_id AND t.account_id = :account_id "
                           "AND s.id < :sid AND s.id > :last_sid "
                           "UNION ALL "
                           "SELECT s.id, c.qty_new, coalesce(l.value/c.qty_new, 0) AS price FROM corp_actions AS c "
                           "LEFT JOIN sequence AS s ON s.type = 5 AND s.operation_id=c.id "
                           "LEFT JOIN ledger AS l ON s.id = l.sid AND l.asset_id=c.asset_id_new AND l.value > 0 "
                           "WHERE c.qty_new > 0 AND c.asset_id_new = :asset_id AND c.account_id = :account_id "
                           "AND s.id < :sid AND s.id > :last_sid "
                           ")ORDER BY id",
                           [(":asset_id", asset_id), (":account_id", account_id),
                            (":sid", seq_id), (":last_sid", last_sid)])
        while query.next():
            deal_sid, deal_qty, deal_price = readSQLrecord(query)  # deal_sid -> trade_sid
            if reminder > 0:
                next_deal_qty = reminder
                reminder = 0
            else:
                next_deal_qty = deal_qty
            if (buy_qty + next_deal_qty) >= qty:  # we are selling less or the same amount as was bought previously
                next_deal_qty = qty - buy_qty
            _ = executeSQL(self.db, "INSERT INTO deals(account_id, asset_id, open_sid, close_sid, qty) "
                                    "VALUES(:account_id, :asset_id, :open_sid, :close_sid, :qty)",
                           [(":account_id", account_id), (":asset_id", asset_id), (":open_sid", deal_sid),
                            (":close_sid", seq_id), (":qty", next_deal_qty)])
            buy_qty = buy_qty + next_deal_qty
            buy_value = buy_value + (next_deal_qty * deal_price)
            if buy_qty == qty:
                break
        # Withdraw value with old quantity of old asset before conversion
        self.appendTransaction(BookAccount.Assets, -buy_qty, -buy_value)
        # Create the same value with new quantity of new asset after conversion
        self.current['asset'] = self.current['peer']
        self.appendTransaction(BookAccount.Assets, self.current['price'], buy_value)

    # Spin-Off is equal to Buy operation with 0 price
    def processAssetEmission(self):
        operation_details = self.current
        # Values for TIMESTAMP, ACCOUNT_ID remains the same
        self.current['asset'] = operation_details['peer']
        self.current['amount'] = operation_details['price']
        self.current['price'] = 0
        self.current['category'] = 0
        self.current['coupon'] = 0
        self.current['peer'] = None
        self.current['fee_tax'] = 0
        self.current['tag'] = 0
        self.processBuy()

    def processCorporateAction(self):
        operationCorpAction = {
            CorporateAction.Merger: self.processAssetConversion,
            CorporateAction.Split: self.processAssetConversion,
            CorporateAction.SymbolChange: self.processAssetConversion,
            CorporateAction.SpinOff: self.processAssetEmission,
            CorporateAction.StockDividend: self.processAssetEmission
        }
        operationCorpAction[self.current['subtype']]()

    # Rebuild transaction sequence and recalculate all amounts
    # timestamp:
    # -1 - re-build from last valid operation (from ledger frontier)
    #      will asks for confirmation if we have more than SILENT_REBUILD_THRESHOLD operations require rebuild
    # 0 - re-build from scratch
    # any - re-build all operations after given timestamp
    def rebuild(self, from_timestamp=-1, fast_and_dirty=False, silent=True):
        operationProcess = {
            TransactionType.Action: self.processAction,
            TransactionType.Dividend: self.processDividend,
            TransactionType.Trade: self.processTrade,
            TransactionType.Transfer: self.processTransfer,
            TransactionType.CorporateAction: self.processCorporateAction
        }

        if from_timestamp >= 0:
            frontier = from_timestamp
            silent = False
        else:
            frontier = self.getCurrentFrontier()
            operations_count = readSQL(self.db, "SELECT COUNT(id) FROM all_transactions WHERE timestamp >= :frontier",
                               [(":frontier", frontier)])
            if operations_count > self.SILENT_REBUILD_THRESHOLD:
                silent = False
                if QMessageBox().warning(None, g_tr('Ledger', "Confirmation"), f"{operations_count}" +
                                         g_tr('Ledger', " operations require rebuild. Do you want to do it right now?"),
                                         QMessageBox.Yes, QMessageBox.No) == QMessageBox.No:
                    return
        if not silent:
            logging.info(g_tr('Ledger', "Re-build ledger from: ") +
                         f"{datetime.fromtimestamp(frontier).strftime('%d/%m/%Y %H:%M:%S')}")
        start_time = datetime.now()
        _ = executeSQL(self.db, "DELETE FROM deals WHERE close_sid >= "
                                "(SELECT coalesce(MIN(id), 0) FROM sequence WHERE timestamp >= :frontier)",
                       [(":frontier", frontier)])
        _ = executeSQL(self.db, "DELETE FROM ledger WHERE timestamp >= :frontier", [(":frontier", frontier)])
        _ = executeSQL(self.db, "DELETE FROM sequence WHERE timestamp >= :frontier", [(":frontier", frontier)])
        _ = executeSQL(self.db, "DELETE FROM ledger_sums WHERE timestamp >= :frontier", [(":frontier", frontier)])
        self.db.commit()

        if fast_and_dirty:  # For 30k operations difference of execution time is - with 0:02:41 / without 0:11:44
            _ = executeSQL(self.db, "PRAGMA synchronous = OFF")
        query = executeSQL(self.db, "SELECT type, id, timestamp, subtype, account, currency, asset, amount, "
                                    "category, price, fee_tax, coupon, peer, tag FROM all_transactions "
                                    "WHERE timestamp >= :frontier", [(":frontier", frontier)])
        while query.next():
            self.current = readSQLrecord(query, named=True)
            seq_query = executeSQL(self.db, "INSERT INTO sequence(timestamp, type, operation_id) "
                                            "VALUES(:timestamp, :type, :operation_id)",
                                   [(":timestamp", self.current['timestamp']), (":type", self.current['type']),
                                    (":operation_id", self.current['id'])])
            self.current_seq = seq_query.lastInsertId()
            operationProcess[self.current['type']]()
            if not silent and (query.at() % 1000) == 0:
                logging.info(g_tr('Ledger', "Processed ") + f"{int(query.at()/1000)}" +
                             g_tr('Ledger', "k records, current frontier: ") +
                             f"{datetime.fromtimestamp(self.current['timestamp']).strftime('%d/%m/%Y %H:%M:%S')}")
        if fast_and_dirty:
            _ = executeSQL(self.db, "PRAGMA synchronous = ON")

        end_time = datetime.now()
        if not silent:
            logging.info(g_tr('Ledger', "Ledger is complete. Elapsed time: ") + f"{end_time - start_time}" +
                         g_tr('Ledger', ", new frontier: ") + f"{datetime.fromtimestamp(self.current['timestamp']).strftime('%d/%m/%Y %H:%M:%S')}")
        self.updateBalancesView()
        self.updateHoldingsView()

    def showRebuildDialog(self, parent):
        rebuild_dialog = RebuildDialog(parent, self.getCurrentFrontier())
        if rebuild_dialog.exec_():
            self.rebuild(from_timestamp=rebuild_dialog.getTimestamp(),
                         fast_and_dirty=rebuild_dialog.isFastAndDirty(), silent=False)


import datetime
import logging
import xlsxwriter
import pandas as pd
from jal.constants import BookAccount, PredefinedAsset, PredefinedCategory, ColumnWidth
from jal.widgets.view_delegate import *
from jal.db.helpers import executeSQL, readSQLrecord
from jal.ui_custom.helpers import g_tr, UseSqlQuery, ConfigureTableView
from jal.reports.helpers import xslxFormat, xlsxWriteRow
from PySide2.QtWidgets import QFileDialog
from PySide2.QtCore import Qt, QObject, Signal, QAbstractTableModel


TREE_LEVEL_SEPARATOR = chr(127)


class ReportType:
    IncomeSpending = 1
    ProfitLoss = 2
    Deals = 3
    ByCategory = 4


#-----------------------------------------------------------------------------------------------------------------------
class PandasModel(QAbstractTableModel):

    CATEGORY_INTEND = "  "

    def __init__(self, data):
        QAbstractTableModel.__init__(self)
        self._data = data

    def rowCount(self, parent=None):
        return self._data.shape[0]

    def columnCount(self, parnet=None):
        return self._data.shape[1] + 1    # +1 as extra leftmost column serves as a category header

    def data(self, index, role=Qt.DisplayRole):
        if index.isValid():
            if role == Qt.DisplayRole:
                if index.column() == 0:
                    row_header = str(self._data.index[index.row()])
                    level = row_header.count(TREE_LEVEL_SEPARATOR)
                    if level > 0:
                        row_header = row_header.rsplit(TREE_LEVEL_SEPARATOR, 1)[1]
                    for i in range(level):
                        row_header = self.CATEGORY_INTEND + row_header
                    return row_header
                else:
                    return self._data.iloc[index.row(), index.column() - 1]
        return None

    def headerData(self, col, orientation, role=Qt.DisplayRole):
        if (orientation == Qt.Horizontal and role == Qt.DisplayRole):
            if col == 0:        # Leftmost column serves as a category header
                return None
            if col == self._data.shape[1]:   # Rightmost total header
                return str(self._data.columns[col-1][1])
            col_date = datetime(year=int(self._data.columns[col-1][1]), month=int(self._data.columns[col-1][2]), day=1)
            return col_date.strftime("%Y %b")
        return None


#-----------------------------------------------------------------------------------------------------------------------
PREPARE_REPORT_QUERY = 0
SHOW_REPORT = 1
REPORT_COLUMNS = 2

class Reports(QObject):
    report_failure = Signal(str)

    def __init__(self, db, report_table_view):
        super().__init__()

        self.db = db
        self.table_view = report_table_view
        self.delegates = []
        self.current_report = None
        self.query = None
        self.dataframe = None
        self.model = None

        self.reports = {
            ReportType.IncomeSpending: (self.prepareIncomeSpendingReport,
                                        self.showPandasReport,
                                        []),
            ReportType.ProfitLoss: (self.prepareProfitLossReport,
                                    self.showSqlQueryReport,
                                    [("period", "Period", ColumnWidth.FOR_DATETIME, None, ReportsYearMonthDelegate),
                                    ("transfer", "In / Out", None, None, ReportsFloat2Delegate),
                                    ("assets", "Assets value", None, None, ReportsFloat2Delegate),
                                    ("result", "Total result", None, None, ReportsFloat2Delegate),
                                    ("profit", "Profit / Loss", None, None, ReportsProfitDelegate),
                                    ("dividend", "Returns", None, None, ReportsFloat2Delegate),
                                    ("tax_fee", "Taxes & Fees", None, None, ReportsFloat2Delegate)]),
            ReportType.Deals: (self.prepareDealsReport,
                               self.showSqlQueryReport,
                               [("asset", "Asset", 300, None, None),
                               ("open_timestamp", "Open Date", ColumnWidth.FOR_DATETIME, None, ReportsTimestampDelegate),
                               ("close_timestamp", "Close Date", ColumnWidth.FOR_DATETIME, None, ReportsTimestampDelegate),
                               ("open_price", "Open Price", None, None, ReportsFloat4Delegate),
                               ("close_price", "Close Price", None, None, ReportsFloat4Delegate),
                               ("qty", "Qty", None, None, ReportsFloatDelegate),
                               ("fee", "Fee", None, None, ReportsFloat2Delegate),
                               ("profit", "P/L", None, None, ReportsProfitDelegate),
                               ("rel_profit", "P/L, %", None, None, ReportsProfitDelegate),
                               ("corp_action", "Note", 200, None, ReportsCorpActionDelegate)]),
            ReportType.ByCategory: (self.prepareCategoryReport,
                                    self.showSqlQueryReport,
                                    [("timestamp", "Timestamp", ColumnWidth.FOR_DATETIME, None, ReportsTimestampDelegate),
                                     ("account", "Account", 200, None, None),
                                     ("name", "Peer Name", 200, None, None),
                                     ("sum", "Amount", 200, None, ReportsFloat2Delegate),
                                     ("note", "Note", -1, None, None)])
        }

    def runReport(self, report_type, begin=0, end=0, account_id=0, group_dates=0):
        if self.reports[report_type][PREPARE_REPORT_QUERY](begin, end, account_id, group_dates):
            self.reports[report_type][SHOW_REPORT](report_type)

    def showSqlQueryReport(self, report_type):
        self.model = UseSqlQuery(self, self.query, self.reports[report_type][REPORT_COLUMNS])
        self.delegates = ConfigureTableView(self.table_view, self.model, self.reports[report_type][REPORT_COLUMNS])
        self.model.select()

    def showPandasReport(self, report_type):
        self.model = PandasModel(self.dataframe)
        self.table_view.setModel(self.model)
        self.delegates = []
        for column in range(self.model.columnCount()):
            if column == 0:
                self.table_view.setColumnWidth(column, 300)
            else:
                self.table_view.setColumnWidth(column, 100)
            self.delegates.append(ReportsPandasDelegate(self.table_view))
            self.table_view.setItemDelegateForColumn(column, self.delegates[-1])
        font = self.table_view.horizontalHeader().font()
        font.setBold(True)
        self.table_view.horizontalHeader().setFont(font)
        self.table_view.show()

    def saveReport(self):
        filename, filter = QFileDialog.getSaveFileName(None, g_tr('Reports', "Save report to:"),
                                                       ".", g_tr('Reports', "Excel files (*.xlsx)"))
        if filename:
            if filter == g_tr('Reports', "Excel files (*.xlsx)") and filename[-5:] != '.xlsx':
                filename = filename + '.xlsx'
        else:
            return
        workbook = xlsxwriter.Workbook(filename=filename)
        formats = xslxFormat(workbook)
        sheet = workbook.add_worksheet(name=g_tr('Reports', "Report"))

        model = self.table_view.model()
        headers = {}
        for col in range(model.columnCount()):
            headers[col] = (model.headerData(col, Qt.Horizontal), formats.ColumnHeader())
        xlsxWriteRow(sheet, 0, headers)

        for row in range(model.rowCount()):
            data_row = {}
            for col in range(model.columnCount()):
                data_row[col] = (model.data(model.index(row, col)), formats.Text(row))
            xlsxWriteRow(sheet, row+1, data_row)

        try:
            workbook.close()
        except:
            logging.error(g_tr('Reports', "Can't save report into file ") + f"'{filename}'")

    def prepareIncomeSpendingReport(self, begin, end, account_id, group_dates):
        _ = executeSQL(self.db, "DELETE FROM t_months")
        _ = executeSQL(self.db, "DELETE FROM t_pivot")
        _ = executeSQL(self.db,
                       "INSERT INTO t_months (month, asset_id, last_timestamp) "
                      "SELECT strftime('%s', datetime(timestamp, 'unixepoch', 'start of month') ) "
                      "AS month, asset_id, MAX(timestamp) AS last_timestamp "
                      "FROM quotes AS q "
                      "LEFT JOIN assets AS a ON q.asset_id=a.id "
                      "WHERE a.type_id=:asset_money "
                      "GROUP BY month, asset_id",
                       [(":asset_money", PredefinedAsset.Money)])
        _ = executeSQL(self.db,
            "INSERT INTO t_pivot (row_key, col_key, value) "
            "SELECT strftime('%s', datetime(t.timestamp, 'unixepoch', 'start of month') ) AS row_key, "
            "t.category_id AS col_key, sum(-t.amount * coalesce(q.quote, 1)) AS value "
            "FROM ledger AS t "
            "LEFT JOIN t_months AS d ON row_key = d.month AND t.asset_id = d.asset_id "
            "LEFT JOIN quotes AS q ON d.last_timestamp = q.timestamp AND t.asset_id = q.asset_id "
            "WHERE (t.book_account=:book_costs OR t.book_account=:book_incomes) "
            "AND t.timestamp>=:begin AND t.timestamp<=:end "
            "GROUP BY row_key, col_key",
            [(":book_costs", BookAccount.Costs), (":book_incomes", BookAccount.Incomes),
             (":begin", begin), (":end", end)])
        self.db.commit()
        self.query = executeSQL(self.db,
                                "SELECT c.id, c.level, c.path, "
                                "strftime('%Y', datetime(p.row_key, 'unixepoch')) AS year, "
                                "strftime('%m', datetime(p.row_key, 'unixepoch')) AS month, p.value "
                                "FROM categories_tree AS c "
                                "LEFT JOIN t_pivot AS p ON p.col_key=c.id "
                                "ORDER BY c.path, year, month")
        table = []
        while self.query.next():
            id, level, category, year, month, value = readSQLrecord(self.query)
            turnover = value if value != '' else 0
            table.append({
                'category': category,
                'Y': year,
                'M': month,
                'turnover': turnover
            })
        data = pd.DataFrame(table)
        data = pd.pivot_table(data, index=['category'], columns=['Y', 'M'], values=['turnover'],
                              aggfunc=sum, fill_value=0.0, margins=True, margins_name=g_tr('Reports', "TOTAL"))
        if data.columns[0][1] == '':   # if some categories have no data and we have null 1st column
            data = data.drop(columns=[data.columns[0]])
        # Calculate sub-totals from bottom to top
        totals = {}
        prev_level = 0
        for index, row in data[::-1].iterrows():
            if index == g_tr('Reports', "TOTAL"):
                continue
            level = index.count(TREE_LEVEL_SEPARATOR)
            if level > prev_level:
                totals[level] = row['turnover']
                prev_level = level
            elif level == prev_level:
                try:
                    totals[level] = totals[level] + row['turnover']
                except KeyError:
                    totals[level] = row['turnover']
            elif level < prev_level:
                try:
                    totals[level] = totals[level] + totals[prev_level] + row['turnover']
                except KeyError:
                    totals[level] = totals[prev_level] + row['turnover']
                sub_total = totals.pop(prev_level, None)
                data.loc[index, :] = sub_total.values
                prev_level = level
        self.dataframe = data
        return True

    def prepareDealsReport(self, begin, end, account_id, group_dates):
        if account_id == 0:
            self.report_failure.emit(g_tr('Reports', "You should select account to create Deals report"))
            return False
        if group_dates == 1:
            self.query = executeSQL(self.db,
                               "SELECT asset, "
                               "strftime('%s', datetime(open_timestamp, 'unixepoch', 'start of day')) as open_timestamp, "
                               "strftime('%s', datetime(close_timestamp, 'unixepoch', 'start of day')) as close_timestamp, "
                               "SUM(open_price*qty)/SUM(qty) as open_price, SUM(close_price*qty)/SUM(qty) AS close_price, "
                               "SUM(qty) as qty, SUM(fee) as fee, SUM(profit) as profit, "
                               "coalesce(100*SUM(qty*(close_price-open_price)-fee)/SUM(qty*open_price), 0) AS rel_profit "
                               "FROM deals_ext "
                               "WHERE account_id=:account_id AND close_timestamp>=:begin AND close_timestamp<=:end "
                               "GROUP BY asset, open_timestamp, close_timestamp "
                               "ORDER BY close_timestamp, open_timestamp",
                               [(":account_id", account_id), (":begin", begin), (":end", end)], forward_only=False)
        else:
            self.query = executeSQL(self.db, "SELECT asset, open_timestamp, close_timestamp, open_price, close_price, "
                                        "qty, fee, profit, rel_profit, corp_action FROM deals_ext "
                                        "WHERE account_id=:account_id AND close_timestamp>=:begin AND close_timestamp<=:end "
                                        "ORDER BY close_timestamp, open_timestamp",
                               [(":account_id", account_id), (":begin", begin), (":end", end)], forward_only=False)
        return True

    def prepareProfitLossReport(self, begin, end, account_id, group_dates):
        if account_id == 0:
            self.report_failure.emit(g_tr('Reports', "You should select account to create Profit/Loss report"))
            return False
        _ = executeSQL(self.db, "DELETE FROM t_months")
        _ = executeSQL(self.db, "INSERT INTO t_months(asset_id, month, last_timestamp) "
                                "SELECT DISTINCT(l.asset_id) AS asset_id, m.m_start, MAX(q.timestamp) AS last_timestamp "
                                "FROM ledger AS l "
                                "LEFT JOIN "
                                "(WITH RECURSIVE months(m_start) AS "
                                "( "
                                "  VALUES(CAST(strftime('%s', date(:begin, 'unixepoch', 'start of month')) AS INTEGER)) "
                                "  UNION ALL "
                                "  SELECT CAST(strftime('%s', date(m_start, 'unixepoch', '+1 month')) AS INTEGER) "
                                "  FROM months "
                                "  WHERE m_start < :end "
                                ") "
                                "SELECT m_start FROM months) AS m "
                                "LEFT JOIN quotes AS q ON q.timestamp<=m.m_start AND q.asset_id=l.asset_id "
                                "WHERE l.timestamp>=:begin AND l.timestamp<=:end AND l.account_id=:account_id "
                                "GROUP BY m.m_start, l.asset_id "
                                "ORDER BY m.m_start, l.asset_id",
                       [(":account_id", account_id), (":begin", begin), (":end", end)])
        self.db.commit()
        self.query = executeSQL(self.db,
            "SELECT DISTINCT(m.month) AS period, coalesce(t.transfer, 0) AS transfer, coalesce(a.assets, 0) AS assets, "
            "coalesce(p.result, 0) AS result, coalesce(o.profit, 0) AS profit, coalesce(d.dividend, 0) AS dividend, "
            "coalesce(f.tax_fee, 0) AS tax_fee "
            "FROM t_months AS m "
            "LEFT JOIN ( "
            "  SELECT mt.month, SUM(-l.amount) AS transfer "
            "  FROM t_months AS mt "
            "  LEFT JOIN ledger AS l ON mt.month = "
            "  CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) "
            "  AND mt.asset_id=l.asset_id "
            "  WHERE l.book_account=:book_transfers AND l.account_id=:account_id GROUP BY mt.month "
            ") AS t ON t.month = m.month "
            "LEFT JOIN ( "
            "  SELECT ma.month, SUM(l.amount*q.quote) AS assets "
            "  FROM t_months AS ma "
            "  LEFT JOIN ledger AS l ON l.timestamp<=ma.month AND l.asset_id=ma.asset_id "
            "  LEFT JOIN quotes AS q ON ma.last_timestamp=q.timestamp AND ma.asset_id=q.asset_id "
            "  WHERE l.account_id =:account_id AND (l.book_account=:book_money OR l.book_account=:book_assets) "
            "  GROUP BY ma.month "
            ") AS a ON a.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) AS month,"
            "  SUM(-l.amount) as result"
            "  FROM ledger AS l  "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS p ON p.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) "
            "  AS INTEGER) AS month, SUM(-l.amount) as profit "
            "  FROM ledger AS l "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) "
            "  AND category_id=:category_profit AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS o ON o.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) AS INTEGER) "
            "  AS month, SUM(-l.amount) as dividend "
            "  FROM ledger AS l "
            "  WHERE (l.book_account=:book_costs OR l.book_account=:book_incomes) "
            "  AND (l.category_id=:category_dividend OR l.category_id=:category_interest) AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS d ON d.month = m.month "
            "LEFT JOIN ( "
            "  SELECT CAST(strftime('%s', date(l.timestamp, 'unixepoch', 'start of month')) "
            "  AS INTEGER) AS month, SUM(-l.amount) as tax_fee "
            "  FROM ledger AS l "
            "  WHERE l.book_account=:book_costs AND l.category_id<>7 AND l.category_id<>8 AND l.account_id=:account_id "
            "  GROUP BY month "
            ") AS f ON f.month = m.month",
            [(":account_id", account_id), (":book_costs", BookAccount.Costs), (":book_incomes", BookAccount.Incomes),
             (":book_money", BookAccount.Money), (":book_assets", BookAccount.Assets),
             (":book_transfers", BookAccount.Transfers), (":category_profit", PredefinedCategory.Profit),
             (":category_dividend", PredefinedCategory.Dividends), (":category_interest", PredefinedCategory.Interest)],
                           forward_only=False)
        return True

    def prepareCategoryReport(self, begin, end, category_id, group_dates):
        if category_id == 0:
            self.report_failure.emit(g_tr('Reports', "You should select category to create By Category report"))
            return False
        self.query = executeSQL(self.db, "SELECT a.timestamp, ac.name AS account, p.name, d.sum, d.note "
                                         "FROM actions AS a "
                                         "LEFT JOIN action_details AS d ON d.pid=a.id "
                                         "LEFT JOIN agents AS p ON p.id=a.peer_id "
                                         "LEFT JOIN accounts AS ac ON ac.id=a.account_id "
                                         "WHERE a.timestamp>=:begin AND a.timestamp<=:end "
                                         "AND d.category_id=:category_id",
                                [(":category_id", category_id), (":begin", begin), (":end", end)], forward_only=False)
        return True


from abc import ABC, abstractmethod

from PySide2.QtCore import Qt, Signal, Property, Slot, QModelIndex
from PySide2.QtWidgets import QWidget, QHBoxLayout, QLineEdit, QLabel, QPushButton, QCompleter

from jal.ui_custom.helpers import g_tr
from jal.ui_custom.reference_data import ReferenceDataDialog, ReferenceTreeDelegate, \
    ReferenceIntDelegate, ReferenceBoolDelegate, ReferenceTimestampDelegate, ReferenceLookupDelegate


# To solve metaclass conflict
class SelectorMeta(type(ABC), type(QWidget)):
    pass


class AbstractReferenceSelector(ABC, QWidget, metaclass=SelectorMeta):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.completer = None
        self.p_selected_id = 0

        self.layout = QHBoxLayout()
        self.layout.setMargin(0)
        self.name = QLineEdit()
        self.name.setText("")
        self.layout.addWidget(self.name)
        self.details = QLabel()
        self.details.setText("")
        self.details.setVisible(False)
        self.layout.addWidget(self.details)
        self.button = QPushButton("...")
        self.button.setFixedWidth(self.button.fontMetrics().width(" ... "))
        self.layout.addWidget(self.button)
        self.setLayout(self.layout)

        self.setFocusProxy(self.name)

        self.button.clicked.connect(self.on_button_clicked)

        self.table = None
        self.selector_field = None
        self.details_field = None
        self.dialog = None

    def getId(self):
        return self.p_selected_id

    def setId(self, selected_id):
        if self.p_selected_id == selected_id:
            return
        self.p_selected_id = selected_id
        self.dialog.Model.setFilter(f"{self.table}.id={selected_id}")
        row_idx = self.dialog.Model.index(0, 0).row()
        name = self.dialog.Model.record(row_idx).value(self.dialog.Model.fieldIndex(self.selector_field))
        self.name.setText(name)
        if self.details_field:
            details = self.dialog.Model.record(row_idx).value(self.dialog.Model.fieldIndex(self.details_field))
            self.details.setText(details)
        self.dialog.Model.setFilter("")
        self.changed.emit()

    @Signal
    def changed(self):
        pass

    selected_id = Property(int, getId, setId, notify=changed, user=True)

    def on_button_clicked(self):
        ref_point = self.mapToGlobal(self.name.geometry().bottomLeft())
        self.dialog.setGeometry(ref_point.x(), ref_point.y(), self.dialog.width(), self.dialog.height())
        res = self.dialog.exec_(enable_selection=True)
        if res:
            self.selected_id = self.dialog.selected_id

    @abstractmethod
    def init_db(self, db):
        pass

    def init_db(self, table, selector_field, details_field=None):
        self.table = table
        self.selector_field = selector_field
        if details_field:
            self.name.setFixedWidth(self.name.fontMetrics().width("X") * 15)
            self.details_field = details_field
            self.details.setVisible(True)
        self.completer = QCompleter(self.dialog.Model)
        self.completer.setCompletionColumn(self.dialog.Model.fieldIndex(self.selector_field))
        self.completer.setCaseSensitivity(Qt.CaseInsensitive)
        self.name.setCompleter(self.completer)
        self.completer.activated[QModelIndex].connect(self.on_completion)

    @Slot(QModelIndex)
    def on_completion(self, index):
        model = index.model()
        self.selected_id = model.data(model.index(index.row(), 0), Qt.DisplayRole)

    def isCustom(self):
        return True


class AccountSelector(AbstractReferenceSelector):
    def __init__(self, parent=None):
        AbstractReferenceSelector.__init__(self, parent)

    def init_db(self, db):
        self.dialog = ReferenceDataDialog(db, "accounts",
                                          [("id", None, 0, None, None),
                                           ("name", g_tr('AccountSelector', "Name"), -1, Qt.AscendingOrder, None),
                                           ("type_id", None, 0, None, None),
                                           ("currency_id", "Currency", None, None, ReferenceLookupDelegate),
                                           ("active", "Act", 32, None, ReferenceBoolDelegate),
                                           ("number", "Account #", None, None, None),
                                           ("reconciled_on", "Reconciled @",
                                            self.fontMetrics().width("00/00/0000 00:00:00") * 1.1,
                                            None, ReferenceTimestampDelegate),
                                           ("organization_id", "Bank", None, None, ReferenceLookupDelegate)],
                                          title=g_tr('AccountSelector', "Accounts"), search_field="full_name", toggle=("active", "Show inactive"),
                                          relations=[("type_id", "account_types", "id", "name", "Account type:"),
                                                     ("currency_id", "currencies", "id", "name", None),
                                                     ("organization_id", "agents", "id", "name", None)])
        super().init_db("accounts", "name")


class AssetSelector(AbstractReferenceSelector):
    def __init__(self, parent=None):
        AbstractReferenceSelector.__init__(self, parent)

    def init_db(self, db):
        self.dialog = ReferenceDataDialog(db, "assets",
                                          [("id", None, 0, None, None),
                                           ("name", "Symbol", None, Qt.AscendingOrder, None),
                                           ("type_id", None, 0, None, None),
                                           ("full_name", "Name", -1, None, None),
                                           ("isin", "ISIN", None, None, None),
                                           ("web_id", "WebID", None, None, None),
                                           ("src_id", "Data source", None, None, ReferenceLookupDelegate)],
                                          title="Assets", search_field="full_name",
                                          relations=[("type_id", "asset_types", "id", "name", "Asset type:"),
                                                     ("src_id", "data_sources", "id", "name", None)])
        super().init_db("assets", "name", "full_name")


class PeerSelector(AbstractReferenceSelector):
    def __init__(self, parent=None):
        AbstractReferenceSelector.__init__(self, parent)

    def init_db(self, db):
        self.dialog = ReferenceDataDialog(db, "agents_ext",
                                          [("id", " ", 16, None, ReferenceTreeDelegate),
                                           ("pid", None, 0, None, None),
                                           ("name", "Name", -1, Qt.AscendingOrder, None),
                                           ("location", "Location", None, None, None),
                                           ("actions_count", "Docs count", None, None, ReferenceIntDelegate),
                                           ("children_count", None, None, None, None)],
                                          title="Choose peer", search_field="name", tree_view=True)
        super().init_db("agents_ext", "name")


class CategorySelector(AbstractReferenceSelector):
    def __init__(self, parent=None):
        AbstractReferenceSelector.__init__(self, parent)

    def init_db(self, db):
        self.dialog = ReferenceDataDialog(db, "categories_ext",
                                          [("id", " ", 16, None, ReferenceTreeDelegate),
                                           ("pid", None, 0, None, None),
                                           ("name", "Name", -1, Qt.AscendingOrder, None),
                                           ("often", "Often", None, None, ReferenceBoolDelegate),
                                           ("special", None, 0, None, None),
                                           ("children_count", None, None, None, None)],
                                          title="Choose category", search_field="name", tree_view=True)
        super().init_db("categories_ext", "name")


class TagSelector(AbstractReferenceSelector):
    def __init__(self, parent=None):
        AbstractReferenceSelector.__init__(self, parent)

    def init_db(self, db):
        self.dialog = ReferenceDataDialog(db, "tags",
                                          [("id", None, 0, None, None),
                                           ("tag", "Tag", -1, Qt.AscendingOrder, None)],
                                          title="Choose tag", search_field="tag")
        super().init_db("tags", "tag")



from jal.widgets.view_delegate import *
from jal.constants import ColumnWidth
from PySide2 import QtWidgets
from PySide2.QtCore import QObject, SIGNAL, Slot
from functools import partial
from jal.widgets.mapper_delegate import MapperDelegate
from jal.ui_custom.helpers import g_tr, UseSqlTable, ConfigureTableView, ConfigureDataMappers
import jal.ui_custom.reference_data as ui               # Full import due to "cyclic" reference


class TableViewConfig:
    BALANCES = 0
    HOLDINGS = 1
    OPERATIONS = 2
    ACTIONS = 3
    ACTION_DETAILS = 4
    TRADES = 5
    DIVIDENDS = 6
    TRANSFERS = 7
    CORP_ACTIONS = 8
    ACCOUNT_TYPES = 9
    ACCOUNTS = 10
    ASSETS = 11
    PEERS = 12
    CATEGORIES = 13
    TAGS = 14
    COUNTRIES = 15

    ACTION_SRC = 0
    ACTION_SIGNAL = 1
    ACTION_SLOT = 2

    DLG_TABLE = 0
    DLG_TITLE = 1
    DLG_COLUMNS = 2
    DLG_SEARCH = 3
    DLG_TOGGLE = 4
    DLG_TREE = 5
    DLG_RELATIONS = 6

    table_names = {
        BALANCES: 'balances',
        HOLDINGS: 'holdings',
        OPERATIONS: 'all_operations',
        ACTIONS: 'actions',
        ACTION_DETAILS: 'action_details',
        TRADES: 'trades',
        DIVIDENDS: 'dividends',
        TRANSFERS: 'transfers_combined',
        CORP_ACTIONS: 'corp_actions'
    }

    table_relations = {
        BALANCES: None,
        HOLDINGS: None,
        OPERATIONS: None,
        ACTIONS: None,
        ACTION_DETAILS: [("category_id", "categories", "id", "name", None),
                         ("tag_id", "tags", "id", "tag", None)],
        TRADES: None,
        DIVIDENDS: None,
        TRANSFERS: None,
        CORP_ACTIONS: None
    }

    table_view_columns = {
        BALANCES: [("level1", None, None, None, None),
                   ("level2", None, None, None, None),
                   ("account_name", g_tr('TableViewConfig', "Account"), ColumnWidth.STRETCH, None, BalanceAccountDelegate),
                   ("balance", g_tr('TableViewConfig', "Balance"), 100, None, BalanceAmountDelegate),
                   ("currency_name", " ", 35, None, BalanceCurrencyDelegate),
                   ("balance_adj", g_tr('TableViewConfig', "Balance, RUB"), 110, None, BalanceAmountAdjustedDelegate),
                   ("days_unreconciled", None, None, None, None),
                   ("active", None, None, None, None)],
        HOLDINGS: [("level1", None, None, None, None),
                   ("level2", None, None, None, None),
                   ("currency", None, None, None, None),
                   ("account", g_tr('TableViewConfig', "C/A"), 32, None, HoldingsAccountDelegate),
                   ("asset", " ", None, None, None),
                   ("asset_name", g_tr('TableViewConfig', "Asset"), ColumnWidth.STRETCH, None, None),
                   ("qty", g_tr('TableViewConfig', "Qty"), None, None, HoldingsFloatDelegate),
                   ("open", g_tr('TableViewConfig', "Open"), None, None, HoldingsFloat4Delegate),
                   ("quote", g_tr('TableViewConfig', "Last"), None, None, HoldingsFloat4Delegate),
                   ("share", g_tr('TableViewConfig', "Share, %"), None, None, HoldingsFloat2Delegate),
                   ("profit_rel", g_tr('TableViewConfig', "P/L, %"), None, None, HoldingsProfitDelegate),
                   ("profit", g_tr('TableViewConfig', "P/L"), None, None, HoldingsProfitDelegate),
                   ("value", g_tr('TableViewConfig', "Value"), None, None, HoldingsFloat2Delegate),
                   ("value_adj", g_tr('TableViewConfig', "Value, RUB"), None, None, HoldingsFloat2Delegate)],
        OPERATIONS: [("type", " ", 10, None, OperationsTypeDelegate),
                     ("id", None, None, None, None),
                     ("timestamp", g_tr('TableViewConfig', "Timestamp"), 150, None, OperationsTimestampDelegate),
                     ("account_id", None, None, None, None),
                     ("account", g_tr('TableViewConfig', "Account"), 300, None, OperationsAccountDelegate),
                     ("num_peer", None, None, None, None),
                     ("asset_id", None, None, None, None),
                     ("asset", None, None, None, None),
                     ("asset_name", None, None, None, None),
                     ("note", g_tr('TableViewConfig', "Notes"), ColumnWidth.STRETCH, None, OperationsNotesDelegate),
                     ("note2", None, None, None, None),
                     ("amount", g_tr('TableViewConfig', "Amount"), None, None, OperationsAmountDelegate),
                     ("qty_trid", None, None, None, None),
                     ("price", None, None, None, None),
                     ("fee_tax", None, None, None, None),
                     ("t_amount", g_tr('TableViewConfig', "Balance"), None, None, OperationsTotalsDelegate),
                     ("t_qty", None, None, None, None),
                     ("currency", g_tr('TableViewConfig', "Currency"), None, None, OperationsCurrencyDelegate),
                     ("reconciled", None, None, None, None)],
        ACTIONS: [],
        ACTION_DETAILS: [("id", None, None, None, None),
                         ("pid", None, None, None, None),
                         ("category_id", g_tr('TableViewConfig', "Category"), 200, None, MapperDelegate),
                         ("tag_id", g_tr('TableViewConfig', "Tag"), 200, None, MapperDelegate),
                         ("sum", g_tr('TableViewConfig', "Amount"), 100, None, MapperDelegate),
                         ("alt_sum", g_tr('TableViewConfig', "Amount *"), 100, None, MapperDelegate),
                         ("note", g_tr('TableViewConfig', "Note"), ColumnWidth.STRETCH, None, None)],
        TRADES: [],
        DIVIDENDS: [],
        TRANSFERS: [],
        CORP_ACTIONS: []
    }

    def __init__(self, parent):
        self.parent = parent
        self.delegates_storage = []   #  Keep references to all created delegates here
        self.views = {
            self.BALANCES: parent.BalancesTableView,
            self.HOLDINGS: parent.HoldingsTableView,
            self.OPERATIONS: parent.OperationsTableView,
            self.ACTIONS: None,
            self.ACTION_DETAILS: parent.ActionDetailsTableView,
            self.TRADES: None,
            self.DIVIDENDS: None,
            self.TRANSFERS: None,
            self.CORP_ACTIONS: None
        }
        self.models = {}
        self.mappers = {}           # Here mapper objects will be stored
        self.widget_mappers = {
            self.BALANCES: None,
            self.HOLDINGS: None,
            self.OPERATIONS: None,
            self.ACTIONS: [("timestamp", parent.ActionTimestampEdit),
                           ("account_id", parent.ActionAccountWidget),
                           ("peer_id", parent.ActionPeerWidget)],
            self.ACTION_DETAILS: None,
            self.TRADES: [("timestamp", parent.TradeTimestampEdit),
                          ("account_id", parent.TradeAccountWidget),
                          ("asset_id", parent.TradeAssetWidget),
                          ("settlement", parent.TradeSettlementEdit),
                          ("number", parent.TradeNumberEdit),
                          ("price", parent.TradePriceEdit) ,
                          ("qty", parent.TradeQtyEdit) ,
                          ("coupon", parent.TradeCouponEdit) ,
                          ("fee", parent.TradeFeeEdit) ],
            self.DIVIDENDS: [("timestamp", parent.DividendTimestampEdit),
                             ("account_id", parent.DividendAccountWidget),
                             ("asset_id", parent.DividendAssetWidget),
                             ("number", parent.DividendNumberEdit),
                             ("sum", parent.DividendSumEdit),
                             ("note", parent.DividendSumDescription),
                             ("sum_tax", parent.DividendTaxEdit),
                             ("tax_country_id", parent.DividendTaxCountry)],
            self.TRANSFERS: [("from_acc_id", parent.TransferFromAccountWidget),
                             ("to_acc_id", parent.TransferToAccountWidget),
                             ("fee_acc_id", parent.TransferFeeAccountWidget),
                             ("from_timestamp", parent.TransferFromTimestamp),
                             ("to_timestamp", parent.TransferToTimestamp),
                             ("fee_timestamp", parent.TransferFeeTimestamp),
                             ("from_amount", parent.TransferFromAmount),
                             ("to_amount", parent.TransferToAmount),
                             ("fee_amount", parent.TransferFeeAmount),
                             ("note", parent.TransferNote)],
            self.CORP_ACTIONS: [("timestamp", parent.CorpActionTimestampEdit),
                                ("number", parent.CorpActionNumberEdit),
                                ("type", parent.CorpActionTypeCombo),
                                ("account_id", parent.CorpActionAccountWidget),
                                ("asset_id", parent.CorpActionAssetBefore),
                                ("asset_id_new", parent.CorpActionAssetAfter),
                                ("qty", parent.CorpActionQtyBeforeEdit),
                                ("qty_new", parent.CorpActionQtyAfterEdit),
                                ("note", parent.CorpActionNodeEdit)]
        }
        self.dialogs = {
            # see DLG_ constants for reference
            self.ACCOUNT_TYPES: ('account_types',
                                 g_tr('TableViewConfig', g_tr('TableViewConfig', "Account Types")),
                                 [("id", None, 0, None, None),
                                  ("name", g_tr('TableViewConfig', "Account Type"), ColumnWidth.STRETCH, Qt.AscendingOrder, None)],
                                 None,
                                 None,
                                 False,
                                 None
                                 ),
            self.ACCOUNTS: ('accounts',
                            g_tr('TableViewConfig', "Accounts"),
                            [("id", None, 0, None, None),
                             ("name", g_tr('TableViewConfig', "Name"), ColumnWidth.STRETCH, Qt.AscendingOrder, None),
                             ("type_id", None, 0, None, None),
                             ("currency_id", g_tr('TableViewConfig', "Currency"), None, None, ui.ReferenceLookupDelegate),
                             ("active", g_tr('TableViewConfig', "Act."), 32, None, ui.ReferenceBoolDelegate),
                             ("number", g_tr('TableViewConfig', "Account #"), None, None, None),
                             ("reconciled_on", g_tr('TableViewConfig', "Reconciled @"), ColumnWidth.FOR_DATETIME,
                              None, ui.ReferenceTimestampDelegate),
                             ("organization_id", g_tr('TableViewConfig', "Bank"), None, None, ui.ReferencePeerDelegate)],
                            "name",
                            ("active", g_tr('TableViewConfig', "Show inactive")),
                            False,
                            [("type_id", "account_types", "id", "name", g_tr('TableViewConfig', "Account type:")),
                             ("currency_id", "currencies", "id", "name", None),
                             ("organization_id", "agents", "id", "name", None)]
                            ),
            self.ASSETS: ("assets",
                          g_tr('TableViewConfig', "Assets"),
                          [("id", None, 0, None, None),
                           ("name", g_tr('TableViewConfig', "Symbol"), None, Qt.AscendingOrder, None),
                           ("type_id", None, 0, None, None),
                           ("full_name", g_tr('TableViewConfig', "Name"), ColumnWidth.STRETCH, None, None),
                           ("isin", g_tr('TableViewConfig', "ISIN"), None, None, None),
                           ("web_id", g_tr('TableViewConfig', "WebID"), None, None, None),
                           ("src_id", g_tr('TableViewConfig', "Data source"), None, None, ui.ReferenceLookupDelegate)],
                          "full_name",
                          None,
                          False,
                          [("type_id", "asset_types", "id", "name", g_tr('TableViewConfig', "Asset type:")),
                           ("src_id", "data_sources", "id", "name", None)]
            ),
            self.PEERS: ("agents_ext",
                          g_tr('TableViewConfig', "Peers"),
                          [("id", " ", 16, None, ui.ReferenceTreeDelegate),
                           ("pid", None, 0, None, None),
                           ("name", g_tr('TableViewConfig', "Name"), ColumnWidth.STRETCH, Qt.AscendingOrder, None),
                           ("location", g_tr('TableViewConfig', "Location"), None, None, None),
                           ("actions_count", g_tr('TableViewConfig', "Docs count"), None, None, ui.ReferenceIntDelegate),
                           ("children_count", None, None, None, None)],
                          "name",
                          None,
                          True,
                          None
            ),
            self.CATEGORIES: ("categories_ext",
                              g_tr('TableViewConfig', "Categories"),
                              [("id", " ", 16, None, ui.ReferenceTreeDelegate),
                               ("pid", None, 0, None, None),
                               ("name", g_tr('TableViewConfig', "Name"), ColumnWidth.STRETCH, Qt.AscendingOrder, None),
                               ("often", g_tr('TableViewConfig', "Often"), None, None, ui.ReferenceBoolDelegate),
                               ("special", None, 0, None, None),
                               ("children_count", None, None, None, None)],
                              "name",
                              None,
                              True,
                              None
            ),
            self.TAGS: ("tags",
                        g_tr('TableViewConfig', "Tags"),
                        [("id", None, 0, None, None),
                         ("tag", g_tr('TableViewConfig', "Tag"), ColumnWidth.STRETCH, Qt.AscendingOrder, None)],
                        "tag",
                        None,
                        False,
                        None
            ),
            self.COUNTRIES: ("countries",
                             g_tr('TableViewConfig', "Countries"),
                             [("id", None, 0, None, None),
                              ("name", g_tr('TableViewConfig', "Country"), ColumnWidth.STRETCH, Qt.AscendingOrder, None),
                              ("code", g_tr('TableViewConfig', "Code"), 50, None, None),
                              ("tax_treaty", g_tr('TableViewConfig', "Tax Treaty"), None, None, ui.ReferenceBoolDelegate)],
                             None,
                             None,
                             False,
                             None
            )
        }
        self.actions = [
            (parent.actionExit,             "triggered()",              QtWidgets.QApplication.instance().quit),
            (parent.action_Load_quotes,     "triggered()",              partial(parent.downloader.showQuoteDownloadDialog, parent)),
            (parent.actionImportStatement,  "triggered()",              parent.statements.loadReport),
            (parent.actionImportSlipRU,     "triggered()",              parent.importSlip),
            (parent.actionBackup,           "triggered()",              parent.Backup),
            (parent.actionRestore,          "triggered()",              parent.Restore),
            (parent.action_Re_build_Ledger, "triggered()",              partial(parent.ledger.showRebuildDialog, parent)),
            (parent.actionAccountTypes,     "triggered()",              partial(self.show_dialog, self.ACCOUNT_TYPES)),
            (parent.actionAccounts,         "triggered()",              partial(self.show_dialog, self.ACCOUNTS)),
            (parent.actionAssets,           "triggered()",              partial(self.show_dialog, self.ASSETS)),
            (parent.actionPeers,            "triggered()",              partial(self.show_dialog, self.PEERS)),
            (parent.actionCategories,       "triggered()",              partial(self.show_dialog, self.CATEGORIES)),
            (parent.actionTags,             "triggered()",              partial(self.show_dialog, self.TAGS)),
            (parent.actionCountries,        "triggered()",              partial(self.show_dialog, self.COUNTRIES)),
            (parent.PrepareTaxForms,        "triggered()",              partial(parent.taxes.showTaxesDialog, parent)),
            (parent.BalanceDate,            "dateChanged(QDate)",       parent.onBalanceDateChange),
            (parent.HoldingsDate,           "dateChanged(QDate)",       parent.onHoldingsDateChange),
            (parent.BalancesCurrencyCombo,  "changed(int)",             parent.OnBalanceCurrencyChange),
            (parent.BalancesTableView,      "doubleClicked(QModelIndex)", parent.OnBalanceDoubleClick),
            (parent.HoldingsCurrencyCombo,  "changed(int)",             parent.OnHoldingsCurrencyChange),
            (parent.ReportRangeCombo,       "currentIndexChanged(int)", parent.onReportRangeChange),
            (parent.RunReportBtn,           "clicked()",                parent.onRunReport),
            (parent.SaveReportBtn,          "clicked()",                parent.reports.saveReport),
            (parent.ShowInactiveCheckBox,   "stateChanged(int)",        parent.OnBalanceInactiveChange),
            (parent.DateRangeCombo,         "currentIndexChanged(int)", parent.OnOperationsRangeChange),
            (parent.ChooseAccountBtn,       "changed(int)",             parent.operations.setAccountId),
            (parent.SearchString,           "textChanged(QString)",     parent.OnSearchTextChange),
            (parent.AddActionDetail,        "clicked()",                partial(parent.operations.addOperationChild, TransactionType.Action)),
            (parent.RemoveActionDetail,     "clicked()",                partial(parent.operations.deleteOperationChild, TransactionType.Action)),
            (parent.DeleteOperationBtn,     "clicked()",                parent.operations.deleteOperation),
            (parent.CopyOperationBtn,       "clicked()",                parent.operations.copyOperation),
            (parent.SaveOperationBtn,       "clicked()",                parent.operations.commitOperation),
            (parent.RevertOperationBtn,     "clicked()",                parent.operations.revertOperation)
        ]

    def tr(self, name):
        pass

    def configure(self, i):
        self.models[i] = UseSqlTable(self.parent, self.table_names[i], self.table_view_columns[i],
                            relations=self.table_relations[i])
        if self.views[i]:
            delegates = ConfigureTableView(self.views[i], self.models[i], self.table_view_columns[i])
            self.delegates_storage.append(delegates)
            self.views[i].show()
        if self.widget_mappers[i]:
            self.mappers[i] = ConfigureDataMappers(self.models[i], self.widget_mappers[i], MapperDelegate)
        else:
            self.mappers[i] = None
        self.models[i].select()

    def configure_all(self):
        for table in self.table_names:
            self.configure(table)
        for action in self.actions:
            QObject.connect(action[self.ACTION_SRC], SIGNAL(action[self.ACTION_SIGNAL]), action[self.ACTION_SLOT])

    @Slot()
    def show_dialog(self, dlg_id):
        ui.ReferenceDataDialog(self.parent.db,
                               self.dialogs[dlg_id][self.DLG_TABLE],
                               self.dialogs[dlg_id][self.DLG_COLUMNS],
                               title=self.dialogs[dlg_id][self.DLG_TITLE],
                               search_field=self.dialogs[dlg_id][self.DLG_SEARCH],
                               toggle=self.dialogs[dlg_id][self.DLG_TOGGLE],
                               tree_view=self.dialogs[dlg_id][self.DLG_TREE],
                               relations=self.dialogs[dlg_id][self.DLG_RELATIONS]
                               ).exec_()
